/*
   This ast2infix() method takes an Abstract Syntax Tree
   for Language-11 and converts the tree into a string that
   contains a "C like" syntax, including infix notation for
   the arithmetic and boolean expressions.

   ast2infix() uses knowledge about operator precedence (in infix
   notation) to use the minimun number of parentheses.
*/

// http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
// http://www.cs.princeton.edu/introcs/11precedence/
int precedence(std::string op)
{
   int result = 0;  // "highest" precedence (i.e., constants)

   if ( 0==op.compare("!") )
   {
      result = 2;
   }
   else if ( 0==op.compare("*")
          || 0==op.compare("/")
          || 0==op.compare("%") )
   {
      result = 4;
   }
   else if ( 0==op.compare("+")
          || 0==op.compare("-") )
   {
      result = 5;
   }
   else if ( 0==op.compare("<")
          || 0==op.compare(">")
          || 0==op.compare("<=")
          || 0==op.compare(">=") )
   {
      result = 7;
   }
   else if ( 0==op.compare("==")
          || 0==op.compare("!=") )
   {
      result = 8;
   }
   else if ( 0==op.compare("&&") )
   {
      result = 11;
   }
   else if ( 0==op.compare("||") )
   {
      result = 12;
   }
   else if ( 0==op.compare("var") || 0==op.compare("set") )
   {
      result = 14;
   }

   return result;
}



std::string ast2infix(Tree* tree, int indentation, bool inLine)
{
   std::string indent = "";
   for (int i = 0; i < indentation; i++)
      indent += " ";

   std::string node = tree->getElement();
   std::string result = "";

   if ( tree->degree() == 0 )  // literal or variable
   {
      result += node;
   }
   else if ( 0==node.compare("prog") )
   {
      result += "Program\n{\n";
      indentation += 3;

      // process each definition
      for (int i = 0; i < tree->degree()-1; i++)
      {
         result += ast2infix( tree->getSubTree(i), indentation, false );
         result += "\n";
      }

      // process the final expression
      result += ast2infix( tree->getSubTree(tree->degree()-1), indentation, false );
      result += ";\n";

      result += indent + "}";
   }
   else if ( 0==node.compare("define") )
   {
      std::string name = tree->getSubTree(0)->getElement();

      std::string rhs = tree->getSubTree(1)->getElement();

      if ( 0==rhs.compare("lambda") )  // function definition
      {
         std::string variable = tree->getSubTree(1)->getSubTree(0)->getElement();

         result += name + "(" + variable + ")\n";
         if ( ! (0==tree->getSubTree(1)->getSubTree(1)->getElement().compare("begin")) )
         {
            result += indent + "{\n";
            indentation += 3;
         }
         result += ast2infix( tree->getSubTree(1)->getSubTree(1), indentation, false );
         if ( ! (0==tree->getSubTree(1)->getSubTree(1)->getElement().compare("begin")) )
         {
            result += ";\n" + indent + "}";
         }
         result += "\n";
      }
      else // variable definition
      {
         indentation += 7 + name.length();
         result += "def " + name + " = "
                          + ast2infix( tree->getSubTree(1), indentation, true )
                          + ";\n";
      }
   }
   else if ( 0==node.compare("lambda") )
   {

   }
   else if ( 0==node.compare("apply") )
   {
      std::string name = tree->getSubTree(0)->getElement();

      result += name + "(";
      indentation += 3;
      result += ast2infix( tree->getSubTree(1), indentation, true );
      result += ")";
   }
   else if ( 0==node.compare("begin") )
   {
      result += "{\n";
      indentation += 3;

      // process each expression
      for (int i = 0; i < tree->degree(); i++)
      {
         result += ast2infix( tree->getSubTree(i), indentation, false );
         if ( 0==tree->getSubTree(i)->getElement().compare("begin") )
            result += "\n";
         else
            result += ";\n";
      }

      result += indent + "}";
   }
   else if ( 0==node.compare("var") )
   {
      std::string name = tree->getSubTree(0)->getElement();

      std::string rhs = tree->getSubTree(1)->getElement();

      if ( 0==rhs.compare("lambda") )  // function definition
      {
         std::string variable = tree->getSubTree(1)->getSubTree(0)->getElement();

         result += name + "(" + variable + ")\n";
         if ( ! (0==tree->getSubTree(1)->getSubTree(1)->getElement().compare("begin")) )
         {
            result += indent + "{\n";
            indentation += 3;
         }
         result += ast2infix( tree->getSubTree(1)->getSubTree(1), indentation, false );
         if ( ! (0==tree->getSubTree(1)->getSubTree(1)->getElement().compare("begin")) )
         {
            result += ";\n" + indent + "}";
         }
      }
      else // variable definition
      {
         indentation += 7 + name.length();
         result += "var " + name + " = "
                          + ast2infix( tree->getSubTree(1), indentation, true );
      }
   }
   else if ( 0==node.compare("set") )
   {
      indentation += 3 + tree->getSubTree(0)->getElement().length();
      result += tree->getSubTree(0)->getElement()
                + " = " + ast2infix( tree->getSubTree(1), indentation, true );
   }
   else if ( 0==node.compare("print") )
   {
      result += "print( ";
      result += ast2infix( tree->getSubTree(0), indentation, true );
      result += " )";
   }
   else if ( 0==node.compare("!") )
   {  // unary operator
      if ( tree->getSubTree(0)->degree() == 2 )
         result += node + "(" + ast2infix(tree->getSubTree(0),0,true ) + ")";
      else
         result += node + ast2infix( tree->getSubTree(0),0,true );
   }
   else if ( 0==node.compare("+") ||   // binary operators
             0==node.compare("-") ||
             0==node.compare("*") ||
             0==node.compare("/") ||
             0==node.compare("%") ||
             0==node.compare("^") ||
             0==node.compare("&&") ||
             0==node.compare("||") ||
             0==node.compare("<") ||
             0==node.compare(">") ||
             0==node.compare("<=") ||
             0==node.compare(">=") ||
             0==node.compare("==") ||
             0==node.compare("!=") )
   {
      int middlePrecedence = precedence( node );
      int leftPrecedence = precedence( tree->getSubTree(0)->getElement() );

      if ( leftPrecedence > middlePrecedence )
      {
         result += "(" + ast2infix( tree->getSubTree(0), 0, true ) + ")";
      }
      else
      {
         result += ast2infix( tree->getSubTree(0), 0, true );
      }

      result += " " + node + " ";

      int rightPrecedence = precedence( tree->getSubTree(1)->getElement() );

      if ( rightPrecedence >= middlePrecedence )
      {
         result += "(" + ast2infix( tree->getSubTree(1), 0, true ) + ")";
      }
      else
      {
         result += ast2infix( tree->getSubTree(1), 0, true );
      }
   }
   else  // shouldn't get here
   {
      result += node;
   }

   if (!inLine)
      result = indent + result;

   return result;
}



std::string ast2infix(Tree* tree)
{
   return ast2infix( tree, 0, true );
}
