/*
   This ast2infix() method takes an Abstract Syntax Tree
   for the language of arithmetic and boolean expressions
   and converts it into a string that contains the infix
   version of the expression.

   ast2infix() uses knowledge about operator precedence (in infix
   notation) to use the minimun number of parentheses.
*/

// http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
// http://www.cs.princeton.edu/introcs/11precedence/
int precedence(std::string op)
{
   int result = 0;  // "highest" precedence (i.e., constants)

   if ( 0==op.compare("!") )
   {
      result = 2;
   }
   else if ( 0==op.compare("^") )
   {
      result = 3;
   }
   else if ( 0==op.compare("*")
          || 0==op.compare("/")
          || 0==op.compare("%") )
   {
      result = 4;
   }
   else if ( 0==op.compare("+")
          || 0==op.compare("-") )
   {
      result = 5;
   }
   else if ( 0==op.compare("<")
          || 0==op.compare(">")
          || 0==op.compare("<=")
          || 0==op.compare(">=") )
   {
      result = 7;
   }
   else if ( 0==op.compare("==")
          || 0==op.compare("!=") )
   {
      result = 8;
   }
   else if ( 0==op.compare("&&") )
   {
      result = 11;
   }
   else if ( 0==op.compare("||") )
   {
      result = 12;
   }

   return result;
}




/*
   This ast2infix() method is essentially a in-order
   traversal of the binary tree.
*/
std::string ast2infix(Tree* tree)
{
   std::string result = "";

   if ( tree->degree() == 0 )  // constant term
   {
      result += tree->getElement();
   }
   else if ( tree->degree() == 1 )  // unary operator
   {
      if ( (tree->getSubTree(0))->degree() == 2 )
      {
         result += tree->getElement() + "(" + ast2infix( tree->getSubTree(0) ) + ")";
      }
      else
      {
         result += tree->getElement() + ast2infix( tree->getSubTree(0) );
      }
   }
   else if ( tree->degree() == 2 )  // binary operator
   {
      int middlePrecedence = precedence( tree->getElement() );
      int leftPrecedence = precedence( (tree->getSubTree(0))->getElement() );

      if ( leftPrecedence > middlePrecedence )
      {
         result += "(" + ast2infix( tree->getSubTree(0) ) + ")";
      }
      else
      {
         result += ast2infix( tree->getSubTree(0) );
      }

      result += " " + tree->getElement() + " ";

      int rightPrecedence = precedence( (tree->getSubTree(1))->getElement() );

      if ( rightPrecedence >= middlePrecedence )
      {
         result += "(" + ast2infix( tree->getSubTree(1) ) + ")";
      }
      else
      {
         result += ast2infix( tree->getSubTree(1) );
      }
   }
   else  // operators with 3 or more operands
   {
      int operatorPrecedence = precedence( tree->getElement() );
      int leftPrecedence = precedence( (tree->getSubTree(0))->getElement() );

      if ( leftPrecedence > operatorPrecedence )
      {
         result += "(" + ast2infix( tree->getSubTree(0) ) + ")";
      }
      else
      {
         result += ast2infix( tree->getSubTree(0) );
      }

      for (int i = 1; i < tree->degree(); i++)
      {
         result += " " + tree->getElement() + " ";

         int rightPrecedence = precedence( (tree->getSubTree(i))->getElement() );

         if ( rightPrecedence >= operatorPrecedence )
         {
            result += "(" + ast2infix( tree->getSubTree(i) ) + ")";
         }
         else
         {
            result += ast2infix( tree->getSubTree(i) );
         }
      }
   }

   return result;
}
