/*
   This ast2infix() method takes an Abstract Syntax Tree for
   the language of arithmetic and boolean expressions with
   nested, local variables and converts the tree into a
   string that contains a C like syntax, including infix
   notation for the arithmetic and boolean expressions.

   ast2infix() uses knowledge about operator precedence (in infix
   notation) to use the minimun number of parentheses.
*/

public class AST2infix
{
   public static String ast2infix(BTree tree)
   {
      return ast2infix( tree, 0, true );
   }


   private static String ast2infix(BTree tree, int indentation, boolean inline)
   {
      String indent = "";
      for (int i = 0; i < indentation; i++)
         indent += " ";

      String node = tree.getElement();
      String result = "";

      if ( tree.degree() == 0 )  // literal or variable
      {
         result += node;
      }
      else if ( node.equals("begin") )
      {
         result += "{\n";
         indentation += 3;

         // process the first expression
         result += ast2infix( tree.getLeftTree(), indentation, false );
         if ( tree.getLeftTree().getElement().equals("begin") )
            result += "\n";
         else
            result += ";\n";

         // process the second expression
         result += ast2infix( tree.getRightTree(), indentation, false );
         if ( tree.getRightTree().getElement().equals("begin") )
            result += "\n";
         else
            result += ";\n";

         result += indent + "}";
      }
      else if ( node.equals("var") )
      {
         indentation += 3 + tree.getLeftTree().getElement().length();
         result += tree.getLeftTree().getElement()
                   + " = " + ast2infix( tree.getRightTree(), indentation, true );
      }
      else if ( node.equals("!") )
      {  // unary operator
         if ( tree.getLeftTree().degree() == 2 )
            result += node + "(" + ast2infix(tree.getLeftTree(),0,true ) + ")";
         else
            result += node + ast2infix( tree.getLeftTree(),0,true );
      }
      else if ( node.equals("+") ||   // binary operators
                node.equals("-") ||
                node.equals("*") ||
                node.equals("/") ||
                node.equals("%") ||
                node.equals("^") ||
                node.equals("&&") ||
                node.equals("||") ||
                node.equals("<") ||
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") ||
                node.equals("==") ||
                node.equals("!=") )
      {
         int middlePrecedence = precedence( node );
         int leftPrecedence = precedence( tree.getLeftTree().getElement() );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + ast2infix( tree.getLeftTree(), 0, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.getLeftTree(), 0, true );
         }

         result += " " + node + " ";

         int rightPrecedence = precedence( tree.getRightTree().getElement() );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + ast2infix( tree.getRightTree(), 0, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.getRightTree(), 0, true );
         }
      }
      else  // shouldn't get here
      {
         result += node;
      }

      if (!inline)
         result = indent + result;

      return result;
   }


   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(String op)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      if ( op.equals("!") )
      {
         result = 2;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 11;
      }
      else if ( op.equals("||") )
      {
         result = 12;
      }
      else if ( op.equals("var") )
      {
         result = 14;
      }

      return result;
   }

}//AST2infix