/*
   This ast2infix() method takes an Abstract Syntax Tree
   for Language-13 and converts the tree into a string that
   contains a "C like" syntax, including infix notation for
   the arithmetic and boolean expressions.

   ast2infix() uses knowledge about operator precedence (in infix
   notation) to use the minimun number of parentheses.
*/

public class AST2infix
{
   public static String ast2infix(Tree tree)
   {
      return ast2infix( tree, 0, true );
   }


   private static String ast2infix(Tree tree, int indentation, boolean inline)
   {
      String indent = "";
      for (int i = 0; i < indentation; i++)
         indent += " ";

      String node = tree.getElement();
      String result = "";

      if ( tree.degree() == 0 )  // literal or variable
      {
         result += node;
      }
      else if ( node.equals("prog") )
      {
         result += "Program\n{\n";
         indentation += 3;

         // process each definition
         for (int i = 0; i < tree.degree()-1; i++)
         {
            result += ast2infix( tree.getSubTree(i), indentation, false );
            result += "\n";
         }

         // process the final expression
         result += ast2infix( tree.getSubTree(tree.degree()-1), indentation, false );
         result += ";\n";

         result += indent + "}";
      }
      else if ( node.equals("define") )
      {
         String name = tree.getSubTree(0).getElement();

         String rhs = tree.getSubTree(1).getElement();

         if ( rhs.equals("lambda") )  // function definition
         {
            String variable = tree.getSubTree(1).getSubTree(0).getElement();

            result += name + "(" + variable + ")\n";
            if ( ! tree.getSubTree(1).getSubTree(1).getElement().equals("begin") )
            {
               result += indent + "{\n";
               indentation += 3;
            }
            result += ast2infix( tree.getSubTree(1).getSubTree(1), indentation, false );
            if ( ! tree.getSubTree(1).getSubTree(1).getElement().equals("begin") )
            {
               result += ";\n" + indent + "}";
            }
            result += "\n";
         }
         else // variable definition
         {
            indentation += 7 + name.length();
            result += "def " + name + " = "
                             + ast2infix( tree.getSubTree(1), indentation, true )
                             + ";\n";
         }
      }
      else if ( node.equals("lambda") )  // we are assuming a short "inline" lambda here
      {
         // get the formal parameter
         result += tree.getSubTree(0).getElement();

         // get the body of the function (use a Maple like notation)
         result += "->";
         indentation += tree.getSubTree(0).getElement().length() + 2;
         Tree expr = tree.getSubTree(1);
         result += ast2infix( expr, indentation, true );
      }
      else if ( node.equals("apply") )
      {
         if ( tree.getSubTree(0).getElement().equals("lambda") )
         { // the anonymous lambda expression case
            result += "("+ast2infix( tree.getSubTree(0), indentation, true )+")";
         }
         if ( tree.getSubTree(0).getElement().equals("apply") )
         { // the anonymous returned function case
            result += ast2infix( tree.getSubTree(0), indentation, true );
         }
         else // the named function case
         {
            String name = tree.getSubTree(0).getElement();
            result += name;
         }

         indentation += 3;
         // get the actual parameter
         result += "(";
         indentation += 3;
         result += ast2infix( tree.getSubTree(1), indentation, true );
         result += ")";
      }
      else if ( node.equals("begin") )
      {
         result += "{\n";
         indentation += 3;

         // process each expression
         for (int i = 0; i < tree.degree(); i++)
         {
            result += ast2infix( tree.getSubTree(i), indentation, false );
            if ( tree.getSubTree(i).getElement().equals("begin") )
               result += "\n";
            else
               result += ";\n";
         }

         result += indent + "}";
      }
      else if ( node.equals("var") )
      {
         String name = tree.getSubTree(0).getElement();

         String rhs = tree.getSubTree(1).getElement();

         if ( rhs.equals("lambda") )  // function definition
         {
            String variable = tree.getSubTree(1).getSubTree(0).getElement();

            result += name + "(" + variable + ")\n";
            if ( ! tree.getSubTree(1).getSubTree(1).getElement().equals("begin") )
            {
               result += indent + "{\n";
               indentation += 3;
            }
            result += ast2infix( tree.getSubTree(1).getSubTree(1), indentation, false );
            if ( ! tree.getSubTree(1).getSubTree(1).getElement().equals("begin") )
            {
               result += ";\n" + indent + "}";
            }
         }
         else // variable definition
         {
            indentation += 7 + name.length();
            result += "var " + name + " = "
                             + ast2infix( tree.getSubTree(1), indentation, true );
         }
      }
      else if ( node.equals("set") )
      {
         indentation += 3 + tree.getSubTree(0).getElement().length();
         result += tree.getSubTree(0).getElement()
                   + " = " + ast2infix( tree.getSubTree(1), indentation, true );
      }
      else if ( node.equals("print") )
      {
         result += "print" + "( ";
         result += ast2infix( tree.getSubTree(0), indentation, true );
         result += " )";
      }
      else if ( node.equals("!") )
      {  // unary operator
         if ( tree.getSubTree(0).degree() == 2 )
            result += node + "(" + ast2infix(tree.getSubTree(0),0,true ) + ")";
         else
            result += node + ast2infix( tree.getSubTree(0),0,true );
      }
      else if ( node.equals("+") ||   // binary operators
                node.equals("-") ||
                node.equals("*") ||
                node.equals("/") ||
                node.equals("%") ||
                node.equals("^") ||
                node.equals("&&") ||
                node.equals("||") ||
                node.equals("<") ||
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") ||
                node.equals("==") ||
                node.equals("!=") )
      {
         int middlePrecedence = precedence( node );
         int leftPrecedence = precedence( tree.getSubTree(0).getElement() );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + ast2infix( tree.getSubTree(0), 0, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.getSubTree(0), 0, true );
         }

         result += " " + node + " ";

         int rightPrecedence = precedence( tree.getSubTree(1).getElement() );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + ast2infix( tree.getSubTree(1), 0, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.getSubTree(1), 0, true );
         }
      }
      else  // shouldn't get here
      {
         result += node;
      }

      if (!inline)
         result = indent + result;

      return result;
   }


   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(String op)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      if ( op.equals("!") )
      {
         result = 2;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 11;
      }
      else if ( op.equals("||") )
      {
         result = 12;
      }
      else if ( op.equals("var") || op.equals("set") )
      {
         result = 14;
      }

      return result;
   }

}//AST2infix