/*
   This ast2infix() method takes an Abstract Syntax Tree
   for Language-14t and converts the tree into a string that
   contains a "C like" syntax, including infix notation for
   the arithmetic and boolean expressions.

   ast2infix() uses knowledge about operator precedence (in infix
   notation) to use the minimun number of parentheses.
*/

public class AST2infix_ver2
{
   public static String ast2infix(Tree tree)
   {
      return ast2infixProg( tree, 0, true );
   }


   private static String ast2infixProg(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }
      result += "Program\n{\n";
      result += ast2infixBody(tree, indentation+3 , true);
      result += indent(indentation) + "}";
      return result;
   }//ast2infixProg()


   private static String ast2infixBody(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }
      // process each definition
      for (int i = 0; i < tree.degree(); i++)
      {
         result += ast2infixE( tree.getSubTree(i), indentation, false );
         result += "\n";
      }
      return result;
   }//ast2infixBody()


   private static String ast2infixE(Tree tree, int indentation, boolean inline)
   {
      String result = null;

      if ( tree.getElement().equals("begin") )
      {
         result = ast2infixBegin(tree, indentation, inline);
      }
      else if ( tree.getElement().equals("var") )
      {
         result = ast2infixVar(tree, indentation, inline);
      }
      else
      {
         result = ast2infixN(tree, indentation, inline);
      }

      return result;
   }//ast2infixE()


   private static String ast2infixBegin(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }
      result += "{\n";
      result += ast2infixBody(tree, indentation+3, false);
      result += indent(indentation) + "}";
      return result;
   }//ast2infixBegin()


   private static String ast2infixVar(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      Tree type = tree.getSubTree(0);

      String name = tree.getSubTree(1).getElement();

      if ( type.getElement().equals("signature") )  // function definition
      {
         // output function type and function name
         result += ast2infixSignature( type, indentation, true );
         result += "\n";
         result += indent(indentation);
         result += name;
         // now output the function definition
         if( tree.getSubTree(2).getElement().equals("lambda") )
         {  // function is defined using a lambda expression
            result += ast2infixLambda( tree.getSubTree(2), indentation, true );
         }
         else  // function is not defined using a lambda expression
         {
            result += " = ";
            result += ast2infixN( tree.getSubTree(2), indentation, true );
         }
         result += "\n";
      }
      else // variable definition
      {
         indentation += 7 + name.length();
         result += "var " + name + " : " + type.getElement() + " = "
                          + ast2infixN( tree.getSubTree(2), indentation, true );
      }

      return result;
   }//ast2infixVar()


   private static String ast2infixN(Tree tree, int indentation, boolean inline)
   {
      String result = null;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         result = "";
         if (!inline)
         {
            result += indent(indentation);
         }
         result += tree.getElement();
      }
      else if ( node.equals("set") )
      {
         result = ast2infixSet(tree, indentation, inline);
      }
      else if ( node.equals("while") )
      {
         result = ast2infixWhile(tree, indentation, inline);
      }
      else if ( node.equals("if") )
      {
         result = ast2infixIf(tree, indentation, inline);
      }
      else if ( node.equals("&&")    // boolean operators
             || node.equals("||") )
      {
         result = ast2infixBool(tree, indentation, inline);
      }
      else if ( node.equals("==")    // equality operators
             || node.equals("!=") )
      {
         result = ast2infixEq(tree, indentation, inline);
      }
      else if ( node.equals("print") )
      {
         result = ast2infixPrint(tree, indentation, inline);
      }
      else if ( node.equals("apply") )
      {
         result = ast2infixApply(tree, indentation, inline);
      }
      else if ( node.equals("lambda") )
      {
         result = ast2infixLambdaInline(tree, indentation, inline);
      }
      else  // shouldn't get here
      {

      }

      return result;
   }//ast2infixN()


   private static String ast2infixSet(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      int newIndentation = indentation + 3 + tree.getSubTree(0).getElement().length();
      result += tree.getSubTree(0).getElement()
                + " = " + ast2infixN( tree.getSubTree(1), newIndentation, true );

      return result;
   }//ast2infixSet()


   private static String ast2infixIf(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      result += "if (" + ast2infixN( tree.getSubTree(0), indentation + 7, true ) + ")\n";
      if ( ! tree.getSubTree(1).getElement().equals("begin") )
      {
         result += ast2infixE( tree.getSubTree(1), indentation+3, false ) + "\n";
      }
      else
      {
         result += ast2infixBegin( tree.getSubTree(1), indentation, false ) + "\n";
      }
      result += indent(indentation) + "else\n";
      if ( ! tree.getSubTree(2).getElement().equals("begin") )
      {
         result += ast2infixE( tree.getSubTree(2), indentation+3, false );
      }
      else
      {
         result += ast2infixBegin( tree.getSubTree(2), indentation, false );
      }

      return result;
   }//ast2infixIf()


   private static String ast2infixWhile(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      result += "while (" + ast2infixN( tree.getSubTree(0), indentation+7, true ) + ")\n";
      if ( ! tree.getSubTree(1).getElement().equals("begin") )
      {
         result += ast2infixE( tree.getSubTree(1), indentation+3, false );
      }
      else
      {
         result += ast2infixBegin( tree.getSubTree(1), indentation, false );
      }

      return result;
   }//ast2infixWhile()


   private static String ast2infixBool(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      int operatorPrecedence = precedence( tree );
      int leftPrecedence = precedence( tree.getSubTree(0) );

      if ( leftPrecedence > operatorPrecedence )
      {
         result += "(" + ast2infixN( tree.getSubTree(0), indentation, true ) + ")";
      }
      else
      {
         result += ast2infixN( tree.getSubTree(0), indentation, true );
      }

      for (int i = 1; i < tree.degree(); i++)
      {
         result += " " + tree.getElement() + " ";

         int rightPrecedence = precedence( tree.getSubTree(i) );

         if ( rightPrecedence >= operatorPrecedence )
         {
            result += "(" + ast2infixN( tree.getSubTree(i), indentation, true ) + ")";
         }
         else
         {
            result += ast2infixN( tree.getSubTree(i), indentation, true );
         }
      }

      return result;
   }//ast2infixBool()


   private static String ast2infixEq(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      int middlePrecedence = precedence( tree );
      int leftPrecedence = precedence( tree.getSubTree(0) );

      if ( leftPrecedence > middlePrecedence )
      {
         result += "(" + ast2infixN( tree.getSubTree(0), indentation, true ) + ")";
      }
      else
      {
         result += ast2infixN( tree.getSubTree(0), indentation, true );
      }

      result += " " + tree.getElement() + " ";

      int rightPrecedence = precedence( tree.getSubTree(1) );

      if ( rightPrecedence >= middlePrecedence )
      {
         result += "(" + ast2infixN( tree.getSubTree(1), indentation, true ) + ")";
      }
      else
      {
         result += ast2infixN( tree.getSubTree(1), indentation, true );
      }

      return result;
   }//ast2infixEq()


   private static String ast2infixPrint(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }
      result += "print( ";
      result += ast2infixE( tree.getSubTree(0), indentation+7, true );
      result += " )";
      return result;
   }//ast2infixPrint()


   private static String ast2infixApply(Tree tree, int indentation, boolean inline)
   {
      String result = null;

      String name = tree.getSubTree(0).getElement();
      if ( name.equals("+")
        || name.equals("-")
        || name.equals("neg")
        || name.equals("*")
        || name.equals("/")
        || name.equals("%")
        || name.equals("^")
        || name.equals("==")
        || name.equals("!=")
        || name.equals("<=")
        || name.equals(">=")
        || name.equals("<")
        || name.equals(">")
        || name.equals("&&")
        || name.equals("||")
        || name.equals("!") )
      {
         result = ast2infixBuiltin( tree, indentation, inline );
      }
      else // anonymous lambda application or user defined function application
      {
         result = "";
         if (!inline)
         {
            result += indent(indentation);
         }
         if ( tree.getSubTree(0).getElement().equals("lambda") )
         { // the anonymous lambda expression case
            result += "("+ast2infixLambdaInline( tree.getSubTree(0), indentation, true )+")";
         }
         else // the named function case
         {
            result += ast2infixN( tree.getSubTree(0), indentation, true );
         }

         // get the actual parameter list
         result += "(";
         if ( tree.degree() > 1 )
         {
            result += ast2infixN( tree.getSubTree(1), indentation+3, true );
            for (int i = 2; i < tree.degree(); i++)
            {
               result += ", " + ast2infixN( tree.getSubTree(i), indentation+3, true );
            }
         }
         result += ")";
      }

      return result;
   }//ast2infixApply()


   private static String ast2infixLambdaInline(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      // get the list of parameter types
      Tree types = tree.getSubTree(0);

      // get the formal parameter list
      result += "(";
      for (int i = 1; i < tree.degree()-1; i++)
      {
         result += tree.getSubTree(i).getElement()
                    + " : " + types.getSubTree(i).getElement();
         if ( i < tree.degree()-2 )
         {
            result += ", ";
         }
      }
      result += ") : " + types.getSubTree(0).getElement();

      // get the body of the function
      result += " -> ";
      Tree expr = tree.getSubTree(tree.degree()-1);
      result += ast2infixE( expr, indentation, true );

      return result;
   }//ast2infixLambdaInline()


   private static String ast2infixLambda(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }
      result += "(";
      // get the formal parameter list (skip the signature)
      if ( tree.degree() > 2 )
      {
         result += tree.getSubTree(1).getElement();
         for (int i = 2; i < tree.degree()-1; i++)
         {
            result += ", " + tree.getSubTree(i).getElement();
         }
      }
      result += ")\n";

      // get the body of the function
      Tree expr = tree.getSubTree(tree.degree()-1);
      if ( ! expr.getElement().equals("begin") )
      {
         result += indent(indentation) + "{\n";
         result += ast2infixE( expr, indentation+3, false );
         result += "\n" + indent(indentation) + "}";
      }
      else
      {
         result += ast2infixBegin( expr, indentation, false );
      }

      return result;
   }//ast2infixLambda()


   private static String ast2infixSignature(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      String rtnType = tree.getSubTree(0).getElement();
      if ( rtnType.equals("signature") )
      {
         rtnType = ast2infixSignature( tree.getSubTree(0), 0, true );
      }
      result += "(";
      for (int i = 1; i < tree.degree(); i++)
      {
         if ( tree.getSubTree(i).getElement().equals("signature") )
         {
            result += ast2infixSignature( tree.getSubTree(i), indentation, true );
         }
         else
         {
            result += tree.getSubTree(i).getElement();
         }
         if ( i < tree.degree()-1 )
         {
            result += ", ";
         }
      }
      result += ")";
      result += " => " + rtnType;

      return result;
   }//ast2infixSignature()


   private static String ast2infixBuiltin(Tree tree, int indentation, boolean inline)
   {
      String result = "";
      if (!inline)
      {
         result += indent(indentation);
      }

      String op = tree.getSubTree(0).getElement();

      if ( tree.degree() == 2 )  // unary operator
      {
         if ( op.equals("neg") ) { op = "-"; }

         if ( tree.getSubTree(1).degree() >= 2 )
         {
            result += op + "(" + ast2infixN( tree.getSubTree(1), indentation, true ) + ")";
         }
         else
         {
            result += op + ast2infixN( tree.getSubTree(1), indentation, true );
         }
      }
      else if ( tree.degree() == 3 )  // binary operator
      {
         int middlePrecedence = precedence( tree );
         int leftPrecedence = precedence( tree.getSubTree(1) );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + ast2infixN( tree.getSubTree(1), indentation, true ) + ")";
         }
         else
         {
            result += ast2infixN( tree.getSubTree(1), indentation, true );
         }

         result += " " + op + " ";

         int rightPrecedence = precedence( tree.getSubTree(2) );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + ast2infixN( tree.getSubTree(2), indentation, true ) + ")";
         }
         else
         {
            result += ast2infixN( tree.getSubTree(2), indentation, true );
         }
      }
      else  // operators with 3 or more operands
      {
         int operatorPrecedence = precedence( tree );
         int leftPrecedence = precedence( tree.getSubTree(1) );

         if ( leftPrecedence > operatorPrecedence )
         {
            result += "(" + ast2infixN( tree.getSubTree(1), indentation, true ) + ")";
         }
         else
         {
            result += ast2infixN( tree.getSubTree(1), indentation, true );
         }

         for (int i = 2; i < tree.degree(); i++)
         {
            result += " " + op + " ";

            int rightPrecedence = precedence( tree.getSubTree(i) );

            if ( rightPrecedence >= operatorPrecedence )
            {
               result += "(" + ast2infixN( tree.getSubTree(i), indentation, true ) + ")";
            }
            else
            {
               result += ast2infixN( tree.getSubTree(i), indentation, true );
            }
         }
      }

      return result;
   }//ast2infixA()


   private static String indent(int indentation)
   {
      String indent = "";
      for (int i = 0; i < indentation; i++)
         indent += " ";
      return indent;
   }//indent()


   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(Tree tree)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      String op = null;

      // get the operator from the tree
      if ( tree.getElement().equals("apply") )
      {
         op = tree.getSubTree(0).getElement();  // built-in operator
      }
      else
      {
         op = tree.getElement(); // special form operator
      }

      if ( op.equals("!") )
      {
         result = 2;
      }
      else if ( op.equals("^") )
      {
         result = 3;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 11;
      }
      else if ( op.equals("||") )
      {
         result = 12;
      }
      else if ( op.equals("var") || op.equals("set") )
      {
         result = 14;
      }

      return result;
   }

}//AST2infix_ver2
