/*
   A Environment object holds <variable,value> pairs and
   has a link to its nesting Environment object (its outer scope).

   For this language, in order to implement function parameters,
   we must have a single namespace for all kinds of variables,
   those that name integers, those that name booleans, and those
   that name functions. Therefore, for this language, the "value"
   that can be stored in a <variable,value> pair can have the
   types "int", "bool", or "lambda".
*/

class Environment
{
   private:
     std::vector<std::string> variables;
     std::vector<Value>  values;
     Environment* nestingLink;

   public:

   Environment()
   {
      nestingLink = NULL;
   }

   Environment(Environment* env)
   {
      nestingLink = env;
   }

   /*
      Add a <variable, value> pair to this environment object.
   */
   void add(std::string variable, Value value)
   {
      variables.push_back(variable);
         values.push_back(value);
   }


   /*
      Look up variable in the environment chain and
      return its associated value.

      Throws an Error if variable is not found.
   */
   Value lookUp(std::string variable)
   {
      unsigned int i;
      for (i=0; i < variables.size(); i++)
         if (0==variable.compare(variables[i]))
            break;

      if ( i < variables.size() )
      {
         return values[i];
      }
      else
      {
         if (nestingLink == NULL)
            throw "undefined variable: " + variable;
         else
            return nestingLink->lookUp(variable);
      }
   }


   /*
      Look up variable in the environment chain.
      Return true if the variable is in the chain,
      otherwise return false.
   */
   bool defined(std::string variable)
   {
      unsigned int i;
      for (i=0; i < variables.size(); i++)
         if (0==variable.compare(variables[i]))
            break;

      if ( i < variables.size() )
      {
         return true;
      }
      else
      {
         if (nestingLink == NULL)
            return false;
         else
            return nestingLink->defined(variable);
      }
   }


   /*
      Look up variable in this environment object.
      Return true if the variable is in this object,
      otherwise return false.
   */
   bool definedLocal(std::string variable)
   {
      unsigned int i;
      for (i=0; i < variables.size(); i++)
         if (0==variable.compare(variables[i]))
            break;

      if ( i < variables.size() )
      {
         return true;
      }
      else
      {
         return false;
      }
   }


   /*
      Update the value associated with variable in the environment chain.
      Return true if the update is succesfull,
      return false if variable is not found in the chain.
   */
   bool update(std::string variable, Value value)
   {
      unsigned int i;
      for (i=0; i < variables.size(); i++)
         if (0==variable.compare(variables[i]))
            break;

      if ( i < variables.size() )
      {
         values.assign(i, value);
         return true;
      }
      else
      {
         if (nestingLink == NULL)
         {
            return false;
         }
         else
         {
            return nestingLink->update(variable, value);
         }
      }
   }


   /*
      Use this operator for information purposes.
   */
   friend std::ostream& operator<<(std::ostream& os, const Environment& e)
   {
      if (e.nestingLink == NULL)
      {
         os << "[Global Environment";
      }
      else
      {
         os << "[Nested Environment";
      }

      for (unsigned int i = 0; i < e.variables.size(); i++)
      {
         os << "\n " << e.variables[i] << " = " << e.values[i];
      }
      os << "]";

      if (e.nestingLink != NULL)
      {
         os << "\n||\n\\/\n" << *(e.nestingLink);
      }
      return os;
   }

};//Environment
