/**
   A Environment object holds <variable,value> pairs
   and <variable, tree> pairs, and has a link to its
   nesting Environment object (its outer scope).
*/
import java.util.ArrayList;

public class Environment
{
   // for the variable name space
   private ArrayList<String> variables;
   private ArrayList<Value>  values;

   // for the function name space
   private ArrayList<String> functions;
   private ArrayList<Tree>  trees;

   private Environment nestingLink;


   // Constructors
   public Environment()
   {
      // initialize the variable name space
      variables = new ArrayList<String>();
      values    = new ArrayList<Value>();

      // initialize the function name space
      functions = new ArrayList<String>();
      trees     = new ArrayList<Tree>();

      nestingLink = null;
   }

   public Environment(Environment env)
   {
      // initialize the variable name space
      variables = new ArrayList<String>();
      values    = new ArrayList<Value>();

      // initialize the function name space
      functions = new ArrayList<String>();
      trees     = new ArrayList<Tree>();

      nestingLink = env;
   }

   /**
      Add a <variable, value> pair to the variable name space.
   */
   public void add(String variable, Value value)
   {
      variables.add(variable);
         values.add(value);
   }

   /**
      Look up variable in the environment chain and
      return its associated value.

      Throws an Error if variable is not found.
   */
   public Value lookUp(String variable)
   {
      int i;
      for (i=0; i < variables.size(); i++)
         if (variable.equals(variables.get(i)))
            break;

      if ( i < variables.size() )
      {
         return values.get(i);
      }
      else
      {
         if (nestingLink == null)
            throw new Error("undefined variable: " + variable);
         else
            return nestingLink.lookUp(variable);
      }
   }

   /**
      Look up variable in the environment chain.
      Return true if the variable is in the chain,
      otherwise return false.
   */
   public boolean defined(String variable)
   {
      int i;
      for (i=0; i < variables.size(); i++)
         if (variable.equals(variables.get(i)))
            break;

      if ( i < variables.size() )
      {
         return true;
      }
      else
      {
         if (nestingLink == null)
            return false;
         else
            return nestingLink.defined(variable);
      }
   }

   /**
      Look up variable in this environment object.
      Return true if the variable is in this object,
      otherwise return false.
   */
   public boolean definedLocal(String variable)
   {
      int i;
      for (i=0; i < variables.size(); i++)
         if (variable.equals(variables.get(i)))
            break;

      if ( i < variables.size() )
      {
         return true;
      }
      else
      {
         return false;
      }
   }

   /**
      Update the value associated with variable in the environment chain.
      Return true if the update is succesfull.
      Return false if variable is not found in the chain.
   */
   public boolean update(String variable, Value value)
   {
      int i;
      for (i=0; i < variables.size(); i++)
         if (variable.equals(variables.get(i)))
            break;

      if ( i < variables.size() )
      {
         values.set(i, value);
         return true;
      }
      else
      {
         if (nestingLink == null)
            return false;
         else
            return nestingLink.update(variable, value);
      }
   }


   // The next four methods make up the function
   // name space part of the environment.


   /**
      Add a <variable, tree> pair to the function name space.
   */
   public void addFunc(String function, Tree tree)
   {
      functions.add(function);
          trees.add(tree);
   }

   /**
      Look up a function in the environment chain and
      return an IPEP pair containing its associated Tree.

      Throws an Error if the function is not found.
   */
   public IPEP lookUpFunc(String function)
   {
      int i;
      for (i=0; i < functions.size(); i++)
         if (function.equals(functions.get(i)))
            break;

      if ( i < functions.size() )
      {
         return new IPEP( trees.get(i), this );
      }
      else
      {
         if (nestingLink == null)
            throw new Error("undefined function: " + function);
         else
            return nestingLink.lookUpFunc(function);
      }
   }

   /**
      Look up a function in the environment chain.
      Return true if the function is in the chain,
      otherwise return false.
   */
   public boolean definedFunc(String function)
   {
      int i;
      for (i=0; i < functions.size(); i++)
         if (function.equals(functions.get(i)))
            break;

      if ( i < functions.size() )
      {
         return true;
      }
      else
      {
         if (nestingLink == null)
            return false;
         else
            return nestingLink.definedFunc(function);
      }
   }

   /**
      Look up function in this environment object.
      Return true if the function is in this object,
      otherwise return false.
   */
   public boolean definedLocalFunc(String function)
   {
      int i;
      for (i=0; i < functions.size(); i++)
         if (function.equals(functions.get(i)))
            break;

      if ( i < functions.size() )
      {
         return true;
      }
      else
      {
         return false;
      }
   }


   /**
      Convert the contents of the environment chain into a string.
      This is mainly for debugging purposes.
   */
   public String toString()
   {
      String result = "";

      if (nestingLink == null)
      {
         result += "[Global Environment";
      }
      else
      {
         result += "[Nested Environment";
      }

      for (int i = 0; i < functions.size(); i++)
      {
         result += "\n " + functions.get(i) + " = " + trees.get(i);
      }
      for (int i = 0; i < variables.size(); i++)
      {
         result += "\n " + variables.get(i) + " = " + values.get(i);
      }
      result += "]";

      if (nestingLink == null)
      {
         return result;
      }
      else
      {
         return result + "\n||\n\\/\n" + nestingLink.toString();
      }
   }

}//Environment