/*
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= '(' 'prog' Define* Exp ')'

     Define ::= '(' 'define' VARIABLE NExp ')'
              | '(' 'define' VARIABLE Lambda ')'

     Lambda ::= '(' 'lambda' VARIABLE Exp ')'

        Exp ::= Begin | Var | NExp

      Begin ::= '(' 'begin' Exp  Exp+ ')'

        Var ::= '(' 'var' VARIABLE NExp ')'
              | '(' 'var' VARIABLE Lambda ')'

       NExp ::= Apply | Set | Print | AExp | BExp
              | VARIABLE | INTEGER | BOOLEAN

      Apply ::= '(' 'apply' VARIABLE NExp ')'

        Set ::= '(' 'set' VARIABLE NExp ')'

      Print ::= '(' 'print' NExp ')'

       AExp ::= '(' BinOp NExp NExp ')'

     BinOp  ::= '+' | '-' | '*' | '/' | '%' | '^'

       BExp ::= '(' BoolOp NExp NExp+ ')'
              | '(' '!' NExp ')'
              | '(' RelOp NExp NExp ')'
              | '('  EqOp NExp NExp ')'

     BoolOp ::= '&&' | '||'
     RelOp  ::= '<' | '>' | '<=' | '>='
      EqOp  ::= '==' | '!='

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*


   This C++ version of the interpreter for Language-11 is different from the Java
   interpreter in one significant way. This C++ inplementation allocates all the
   environment objects on the C++ runtime stack. This means that all of these
   objects have "automatic" lifetimes, which means that they are deallocated when
   the function that created them returns. Because of this, this interpreter cannot
   implement Language-12. For Langauge-12, the environment objects must have lifetimes
   that are indefinite.

   We can try to interpret Language-12 programs with this interpreter by removing
   from the evaluateApply() function the check that prevents function returnn values,
   If we do that, this interpreter will crash when it tries to apply a function that
   was a return value, since that function's environment pointer will be a "dangling
   pointer", a pointer that points to deallocated memory.

   In the case of the Java version of this interpreter, we can make the Language-11
   interpreter interpret Language-12 programs by just removing from the evaluateApply()
   function the check that prevents function returnn values, Since Java has a garbage
   collector, all objects have indefinite lifetimes, so the environment for a function
   that is a return value will always be available for when that function needs to be
   applied.
*/
#include <math.h>

Value evaluateProg(Tree* tree);
void evaluateDefine(Tree* tree, Environment* env);
Value evaluateE(Tree* tree, Environment* env);
Value evaluateBegin(Tree* tree, Environment* env);
Value evaluateVar(Tree* tree, Environment* env);
Value evaluateN(Tree* tree, Environment* env);
Value evaluateApply(Tree* tree, Environment* env);
Value evaluateSet(Tree* tree, Environment* env);
Value evaluatePrint(Tree* tree, Environment* env);
Value evaluateA(Tree* tree, Environment* env);
Value evaluateB(Tree* tree, Environment* env);
Value evaluateEq(Tree* tree, Environment* env);
Value evaluateR(Tree* tree, Environment* env);

std::string boolTag = "bool";
std::string  intTag = "int";
std::string lambdaTag = "lambda";


// Global environment for function and variable definitions.
Environment globalEnv;


Value evaluate(Tree* tree)
{
   //globalEnv = new Environment();

   return  evaluateProg( tree );
}//evaluate()


Value evaluateProg(Tree* tree)
{
   Value result;  // use the default Value constructor

   // Evaluate the definitions in the Prog term.
   // Each definition will have the side effect of putting
   // something in one of the global environments.
   for (int i = 0; i < tree->degree()-1; i++)
   {
      evaluateDefine( tree->getSubTree(i), &globalEnv );
   }

   // Evaluate the single expression.
   result = evaluateE( tree->getSubTree(tree->degree()-1), &globalEnv );

   return result;
}//evaluateProg()


// Evaluate definitions. Notice that there is no return value.
// This function only mutates the global environment objects.
void evaluateDefine(Tree* tree, Environment* env)
{
   // get the variable or function name
   std::string name = tree->getSubTree(0)->getElement();

   // get the expression
   Tree* expr = tree->getSubTree(1);

   // determine if the definition is a function
   if (0==expr->getElement().compare("lambda"))
   {  ////(1) create a "lambda" Value object
      IPEP ipep; ipep.ip=expr; ipep.ep=env;
      env->add( name, Value(&lambdaTag, ipep) );
      std::cout << *env << "\n\n"; // for debugging purposes
   }
   else // not a function, so evaluate the expression
   {
      Value value = evaluateE(expr, env);

      // check if this variable has already been defined
      if ( ! env->definedLocal(name) )
      {
         env->add(name, value);
      }
      else // this variable is already in the local Environment object
      {
         env->update(name, value);
      }
      std::cout << *env << "\n\n"; // for debugging purposes
   }
}//evaluateDefine()


/**
   Notice that we evaluate an expression using the included environment which
   is a linked list of environment objects, terminating with the global environment.
*/
Value evaluateE(Tree* tree, Environment* env)
{
   Value result;  // use the default Value constructor

   std::string node = tree->getElement();

   if ( 0==node.compare("begin") )
   {
      result = evaluateBegin(tree, env);
   }
   else if ( 0==node.compare("var") )
   {
      result = evaluateVar(tree, env);
   }
   else
   {
      result = evaluateN(tree, env);
   }

   return result;
}//evaluateE()


Value evaluateBegin(Tree* tree, Environment* env)
{
   // create a new environment "nested" in the previous environment
   Environment localEnv(env);

   Value result;  // use the default Value constructor

   // evaluate each expression (using the nested environment)
   for (int i = 0; i < tree->degree(); i++)
   {
      // Notice that, except for the last expression, this evaluation
      // serves no purpose unless the expression is a "var" or "set" term,
      // which mutate environment objects, or a "begin" term which
      // contains a "var" or "set" term. In other words, unless this
      // expression has a side-effect, it does not do anything.
      result = evaluateE( tree->getSubTree(i), &localEnv );
   }

   // Notice that when this evaluateBegin() method returns, the
   // localEnv object becomes a garbage object, so it is implicitly
   // removed from the list of Environment objects.

   return result;
}//evaluateBegin()


Value evaluateVar(Tree* tree, Environment* env)
{
   Value result;  // use the default Value constructor

   // get the variable
   std::string variable = tree->getSubTree(0)->getElement();

   // get the expression
   Tree* expr = tree->getSubTree(1);

   if (0==expr->getElement().compare("lambda") )
   {
      // check if this variable has already been defined
      if ( ! env->definedLocal(variable) )
      {  ////(2) create a "lambda" Value object
         IPEP ipep; ipep.ip=expr; ipep.ep=env;
         result = Value(&lambdaTag, ipep);
         env->add( variable, result );
      }
      else // this variable is already in the local Environment object
      {
         throw "local function already exisits: " + variable;
      }
      std::cout << *env << "\n\n"; // for debugging purposes
   }
   else // not a function, so evaluate the expression
   {
      result = evaluateN(expr, env);

      // check if this variable has already been defined
      if ( ! env->definedLocal(variable) )
      {
         env->add(variable, result);
      }
      else // this variable is already in the local Environment object
      {
         throw "local variable already exisits: " + variable;
      }
      std::cout << *env << "\n\n"; // for debugging purposes
   }

   return result;

}//evaluateVar()


Value evaluateN(Tree* tree, Environment* env)
{
   Value result;  // use the default Value constructor

   std::string node = tree->getElement();

   if ( tree->degree() == 0 )
   {
      if ( 0==node.compare("true") || 0==node.compare("false") )
      {
         Value r( &boolTag, 0==node.compare("true") );
         result = r;
      }
      else if ( env->defined(node) )
      {
         result = env->lookUp( node );
      }
      else // just assume the node is an integer literal
      {
         Value r( &intTag, atoi(node.c_str()) );
         result = r;
      }
   }
   else if ( 0==node.compare("apply") )
   {
      result = evaluateApply(tree, env);
   }
   else if ( 0==node.compare("set") )
   {
      result = evaluateSet(tree, env);
   }
   else if ( 0==node.compare("print") )
   {
      result = evaluatePrint(tree, env);
   }
   else if ( 0==node.compare("+") ||     // arithmetic operators
             0==node.compare("-") ||
             0==node.compare("*") ||
             0==node.compare("/") ||
             0==node.compare("%") ||
             0==node.compare("^") )
   {
      result = evaluateA(tree, env);
   }
   else if ( 0==node.compare("&&") ||    // boolean operators
             0==node.compare("||") ||
             0==node.compare("!") )
   {
      result = evaluateB(tree, env);
   }
   else if ( 0==node.compare("==") ||    // equality operators
             0==node.compare("!=") )
   {
      result = evaluateEq(tree, env);
   }
   else if ( 0==node.compare("<") ||     // relational operators
             0==node.compare(">") ||
             0==node.compare("<=") ||
             0==node.compare(">=") )
   {
      result = evaluateR(tree, env);
   }
   else
   {
      throw "invalid expression: " + ast2infix(tree);
   }

   return result;
}//evaluateN()


/**
   This method "applies" a function to an actual parameter.
   This method evaluates the body of the function in an environment
   that binds the actual parameter value (from this function application)
   to the formal parameter (from the function's definition).
*/
Value evaluateApply(Tree* tree, Environment* env)
{
   Value result;  // use the default Value constructor

   // Evaluate the second expression (using the original environment).
   Value actualParam = evaluateE( tree->getSubTree(1), env );

   // Retrive, from within the current environment,
   // a reference to the function value.
   Value func = env->lookUp( tree->getSubTree(0)->getElement() );

   // Check that this value has the correct type
   if ( ! 0==func.tag->compare("lambda") )
   {
      throw "variable not a function: " + tree->getSubTree(0)->getElement();
   }

   // Retrive, from within the Value object,
   // a reference to the function's IPEP pair.
   IPEP ipepPair = func.valueL;

   // Retrive, from within the IPEP pair,
   // a reference to the function's definition
   // (i.e., the function's "lambda expression").
   Tree* function = ipepPair.ip;

   // Create a new environment "nested" in the environment
   // pointed to by the ipep pair.
   Environment localEnv(ipepPair.ep);  // NOTE: static scope

   // Bind the actual parameter value to the formal parameter.
   // The binding is done in the new, local environment.
   std::string formalParam = function->getSubTree(0)->getElement();
   localEnv.add(formalParam, actualParam);
   std::cout << localEnv << "\n\n"; // for debugging purposes

   // Finally, evaluate the body of the function using the
   // new environment (which contains the binding of the actual
   // parameter value to the function's formal parameter).
   result = evaluateE( function->getSubTree(1), &localEnv );

   // Make sure that the return value is not a "lambda" value.
//   if ( 0==result.tag->compare("lambda") )
//   {
//      throw "functions as return values not allowed: " + ast2infix(tree);
//   }

   return result;
}//evaluateApply()


Value evaluateSet(Tree* tree, Environment* env)
{
   // get the variable
   std::string variable = tree->getSubTree(0)->getElement();

   // get, and then evaluate, the expression
   Tree* expr = tree->getSubTree(1);
   Value value = evaluateE(expr, env);

   // check if this variable has already been defined
   if ( ! env->defined(variable) )
   {
      throw "undefined variable: " + variable;
   }
   else // this variable is already in the environment
   {
      env->update(variable, value);
   }

   std::cout << *env << "\n\n"; // for debugging purposes

   return value;
}//evaluateSet()


Value evaluatePrint(Tree* tree, Environment* env)
{
   Value result = evaluateN( tree->getSubTree(0), env );

   std::cout << result;

   return result;
}//evaluatePrint()


// Evaluate an arithmetic expression
Value evaluateA(Tree* tree, Environment* env)
{
   int result = 0;

   std::string node = tree->getElement();

   Value valueL = evaluateN( tree->getSubTree(0), env );
   Value valueR = evaluateN( tree->getSubTree(1), env );

   if ( ! 0==valueL.tag->compare("int") )
   {
      throw "illegal arithmetic value: " + *valueL.tag;
   }
   if ( ! 0==valueR.tag->compare("int") )
   {
      throw "illegal arithmetic value: " + *valueR.tag;
   }

   int resultL = valueL.valueI;
   int resultR = valueR.valueI;

   if ( 0==node.compare("+") )
   {
      result = resultL + resultR;
   }
   else if ( 0==node.compare("-") )
   {
      result = resultL - resultR;
   }
   else if ( 0==node.compare("*") )
   {
      result = resultL * resultR;
   }
   else if ( 0==node.compare("/") )
   {
      result = resultL / resultR;
   }
   else if ( 0==node.compare("%") )
   {
      result = resultL % resultR;
   }
   else if ( 0==node.compare("^") )
   {
      result = (int)pow(resultL, resultR);
   }
   else
   {
      throw "invalid arithmetic operator: " + node;
   }

   return Value( &intTag, result );
}//evaluateA()


// Evaluate a boolean expression
Value evaluateB(Tree* tree, Environment* env)
{
   bool result = false;

   std::string node = tree->getElement();

   Value valueL = evaluateN( tree->getSubTree(0), env );
   if ( ! 0==valueL.tag->compare("bool") )
   {
      throw "illegal boolean value: " + *valueL.tag;
   }
   bool resultL = valueL.valueB;

   bool resultR = false;
   if ( tree->degree() == 2 )  // check for second operand
   {
      Value valueR = evaluateN( tree->getSubTree(1), env );
      if ( ! 0==valueR.tag->compare("bool") )
      {
         throw "illegal boolean value: " + *valueR.tag;
      }
      resultR = valueR.valueB;
   }

   if ( 0==node.compare("&&") )  // not short circuited
   {
      result = resultL && resultR;
   }
   else if ( 0==node.compare("||") )  // not short circuited
   {
      result = resultL || resultR;
   }
   else if ( 0==node.compare("!") )  // unary operator
   {
      result = ! resultL;
   }
   else
   {
      throw "invalid boolean operator: " + node;
   }

   return Value( &boolTag, result );
}//evaluateB()


// Evaluate an equality expression (which is a kind of boolean expression)
Value evaluateEq(Tree* tree, Environment* env)
{
   bool result = false;

   std::string node = tree->getElement();

   Value valueL = evaluateE( tree->getSubTree(0), env );
   Value valueR = evaluateE( tree->getSubTree(1), env );

   if ( ! 0==valueL.tag->compare(*(valueR.tag)) )
   {
      result = false;
   }
   else if ( 0==valueL.tag->compare("int") )
   {
      result = (valueL.valueI == valueR.valueI);
   }
   else if ( 0==valueL.tag->compare("bool") )
   {
      result = (valueL.valueB == valueR.valueB);
   }
   else
   {
      throw "invalid Value tag: " + *valueL.tag;
   }

   if ( 0==node.compare("!=") )
   {
      result = ! result;
   }
   else  if ( ! 0==node.compare("==") )
   {
     throw "invalid equality operator: " + node;
   }

   return Value( &boolTag, result );
}//evaluateEq()


// Evaluate a relational expression (which is a kind of boolean expression)
Value evaluateR(Tree* tree, Environment* env)
{
   bool result = false;

   std::string node = tree->getElement();

   Value valueL = evaluateN( tree->getSubTree(0), env );
   Value valueR = evaluateN( tree->getSubTree(1), env );

   if ( ! 0==valueL.tag->compare("int") )
   {
      throw "illegal arithmetic value: " + *valueL.tag;
   }
   if ( ! 0==valueR.tag->compare("int") )
   {
      throw "illegal arithmetic value: " + *valueR.tag;
   }

   int resultL = valueL.valueI;
   int resultR = valueR.valueI;

   if (0== node.compare("<") )
   {
      result = resultL < resultR;
   }
   else if ( 0==node.compare(">") )
   {
      result = resultL > resultR;
   }
   else if ( 0==node.compare("<=") )
   {
      result = resultL <= resultR;
   }
   else if ( 0==node.compare(">=") )
   {
      result = resultL >= resultR;
   }
   else
   {
      throw "invalid relational operator: " + node;
   }

   return Value( &boolTag, result );
}//evaluateR()
