/*
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= '(' 'prog' Define* Exp ')'

     Define ::= '(' 'define' VARIABLE NExp ')'   // now we need only one clause here!

        Exp ::= Begin | Var | NExp

      Begin ::= '(' 'begin' Exp  Exp+ ')'

        Var ::= '(' 'var' VARIABLE NExp ')'      // now we need only one clause here!

       NExp ::= Lambda | Apply | Set | Print | AExp | BExp
              | VARIABLE | INTEGER | BOOLEAN

     Lambda ::= '(' 'lambda' VARIABLE Exp ')'

      Apply ::= '(' 'apply' NExp NExp ')'        // notice a change here

        Set ::= '(' 'set' VARIABLE NExp ')'

      Print ::= '(' 'print' NExp ')'

       AExp ::= '(' BinOp NExp NExp ')'

     BinOp  ::= '+' | '-' | '*' | '/' | '%' | '^'

       BExp ::= '(' BoolOp NExp NExp+ ')'
              | '(' '!' NExp ')'
              | '(' RelOp NExp NExp ')'
              | '('  EqOp NExp NExp ')'

     BoolOp ::= '&&' | '||'
     RelOp  ::= '<' | '>' | '<=' | '>='
      EqOp  ::= '==' | '!='

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*


   This language makes the Lambda term offically into an expression (see
   the NExp term). This is a very important change. It is what makes
   functions, i.e., "lambda expressions", truely into first class objects.
   Because of this change, we can drop an extra clause from each of the Define
   and Var terms (in fact, we could even get rid of the Define term) and we can
   generalize the Apply term. These changes also allow us to make use of
   "anonymous functions".

   In the interpreter, we simplify the definition of evluateDefine() and evaluateVar()
   to account for the removed clause. We change evaluateApply() to account for the
   more generalized NExp. And, most importantly, we now need to write an
   evaluateLambda() function. The evaluateLambda() function will create a Value
   object with the "lambda" tag. Depanding on who called evaluateLambda(), this
   "lambda value" will either get put in an environment, or it might get "applied"
   directly to an argument. The "lambda value" returned by evaluateLambda() is
   truly a "closure".
*/

public class Evaluate
{
   // Global environment for function and variable definitions.
   private static Environment globalEnv;



   public static Value evaluate(Tree tree) throws EvalException
   {
      globalEnv = new Environment();


      return  evaluateProg( tree );
   }//evaluate()


   private static Value evaluateProg(Tree tree) throws EvalException
   {
      Value result = null;

      // Evaluate the definitions in the Prog term.
      // Each definition will have the side effect of putting
      // something in one of the global environments.
      for (int i = 0; i < tree.degree()-1; i++)
      {
         evaluateDefine( tree.getSubTree(i), globalEnv );
      }

      // Evaluate the single expression.
      result = evaluateE( tree.getSubTree(tree.degree()-1), globalEnv );

      return result;
   }//evaluateProg()


   // Evaluate definitions. Notice that there is no return value.
   // This function only mutates the global environment objects.
   private static void evaluateDefine(Tree tree, Environment env) throws EvalException
   {
      // get the variable or function name
      String name = tree.getSubTree(0).getElement();

      // get the expression
      Tree expr = tree.getSubTree(1);

      Value value = evaluateE(expr, env);

      // check if this variable has already been defined
      if ( ! env.definedLocal(name) )
      {
         env.add(name, value);
      }
      else // this variable is already in the local Environment object
      {
         env.update(name, value);
      }
//      System.out.println( env + "\n" ); // for debugging purposes
   }//evaluateDefine()


   /**
      Notice that we evaluate an expression using the included environment which
      is a linked list of environment objects, terminating with the global environment.
   */
   private static Value evaluateE(Tree tree, Environment env) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( node.equals("begin") )
      {
         result = evaluateBegin(tree, env);
      }
      else if ( node.equals("var") )
      {
         result = evaluateVar(tree, env);
      }
      else
      {
         result = evaluateN(tree, env);
      }

      return result;
   }//evaluateE()


   private static Value evaluateBegin(Tree tree, Environment env) throws EvalException
   {
      // create a new environment "nested" in the previous environment
      Environment localEnv = new Environment(env);

      Value result = null;

      // evaluate each expression (using the nested environment)
      for (int i = 0; i < tree.degree(); i++)
      {
         // Notice that, except for the last expression, this evaluation
         // serves no purpose unless the expression is a "var" or "set" term,
         // which mutate environment objects, or a "begin" term which
         // contains a "var" or "set" term. In other words, unless this
         // expression has a side-effect, it does not do anything.
         result = evaluateE( tree.getSubTree(i), localEnv );
      }

      // Notice that when this evaluateBegin() method returns, the
      // localEnv object becomes a garbage object, so it is implicitly
      // removed from the list of Environment objects.

      return result;
   }//evaluateBegin()


   private static Value evaluateVar(Tree tree, Environment env) throws EvalException
   {
      Value result = new Value("bool", false);  // default return value

      // get the variable
      String variable = tree.getSubTree(0).getElement();

      // get the expression
      Tree expr = tree.getSubTree(1);

      result = evaluateN(expr, env);

      // check if this variable has already been defined
      if ( !env.definedLocal(variable) )
      {
         env.add(variable, result);
      }
      else // this variable is already in the local Environment object
      {
         throw new EvalException("local variable already exisits: " + variable);
      }
//      System.out.println( env + "\n" ); // for debugging purposes

      return result;

   }//evaluateVar()


   private static Value evaluateN(Tree tree, Environment env) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( "bool", node.equals("true") );
         }
         else if ( node.matches("^[0-9][0-9]*") )
         {
            result = new Value( "int", Integer.parseInt( node ) );
         }
         else if ( env.defined(node) )
         {
            result = env.lookUp( node );
         }
         else
         {
            throw new EvalException("undefined variable: " + node);
         }
      }
      else if ( node.equals("lambda") )
      {
         result = evaluateLambda(tree, env);
      }
      else if ( node.equals("apply") )
      {
         result = evaluateApply(tree, env);
      }
      else if ( node.equals("set") )
      {
         result = evaluateSet(tree, env);
      }
      else if ( node.equals("print") )
      {
         result = evaluatePrint(tree, env);
      }
      else if ( node.equals("+") ||     // arithmetic operators
                node.equals("-") ||
                node.equals("*") ||
                node.equals("/") ||
                node.equals("%") ||
                node.equals("^") )
      {
         result = evaluateA(tree, env);
      }
      else if ( node.equals("&&") ||    // boolean operators
                node.equals("||") ||
                node.equals("!") )
      {
         result = evaluateB(tree, env);
      }
      else if ( node.equals("==") ||    // equality operators
                node.equals("!=") )
      {
         result = evaluateEq(tree, env);
      }
      else if ( node.equals("<") ||     // relational operators
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") )
      {
         result = evaluateR(tree, env);
      }
      else
      {
         throw new EvalException("invalid expression: " + AST2infix.ast2infix(tree));
      }

      return result;
   }//evaluateN()


   /**

   */
   private static Value evaluateLambda(Tree tree, Environment env) throws EvalException
   {
      Value result;

      result = new Value( "lambda", new IPEP(tree,env) );

      return result;
   }//evaluateLambda()


   /**
      This method "applies" a function to an actual parameter.
      This method evaluates the body of the function in an environment
      that binds the actual parameter value (from this function application)
      to the formal parameter (from the function's definition).
   */
   private static Value evaluateApply(Tree tree, Environment env) throws EvalException
   {
      Value result;

      // Evaluate the second expression (using the original environment).
      Value actualParam = evaluateN( tree.getSubTree(1), env );

      // Evaluate the first expression, which is supposed
      // to evalute to a function value.
      Value func = evaluateN( tree.getSubTree(0), env );

      // Check that this value has the correct type
      if ( ! func.tag.equals("lambda") )
      {
         throw new EvalException("variable not a function: " + tree.getSubTree(0).getElement());
      }

      // Retrive, from within the Value object,
      // a reference to the function's IPEP pair.
      IPEP ipepPair = func.valueL;

      // Retrive, from within the IPEP pair,
      // a reference to the function's definition
      // (i.e., the function's "lambda expression").
      Tree function = ipepPair.exp;

      // Create a new environment "nested" in the environment
      // pointed to by the ipep pair.
      Environment localEnv = new Environment(ipepPair.env);  // NOTE: static scope

      // Bind the actual parameter value to the formal parameter.
      // The binding is done in the new, local environment.
      String formalParam = function.getSubTree(0).getElement();
      localEnv.add(formalParam, actualParam);
//      System.out.println( localEnv + "\n" ); // for debugging purposes

      // Finally, evaluate the body of the function using the
      // new environment (which contains the binding of the actual
      // parameter value to the function's formal parameter).
      result = evaluateE( function.getSubTree(1), localEnv );

      // Make sure that the return value is not a "lambda" value.
      //if ( result.tag.equals("lambda") )
      //{
      //   throw new EvalException("functions as return values not allowed: " + result);
      //}

      return result;
   }//evaluateApply()


   private static Value evaluateSet(Tree tree, Environment env) throws EvalException
   {
      // get the variable
      String variable = tree.getSubTree(0).getElement();

      // get, and then evaluate, the expression
      Tree expr = tree.getSubTree(1);
      Value value = evaluateE(expr, env);

      // check if this variable has already been defined
      if ( ! env.defined(variable) )
      {
         throw new EvalException("undefined variable: " + variable);
      }
      else // this variable is already in the environment
      {
         env.update(variable, value);
      }

//      System.out.println( env + "\n" ); // for debugging purposes

      return value;

   }//evaluateSet()


   private static Value evaluatePrint(Tree tree, Environment env) throws EvalException
   {
      Value result = evaluateE( tree.getSubTree(0), env );

      System.out.println( result );

      return result;
   }//evaluatePrint()


   // Evaluate an arithmetic expression
   private static Value evaluateA(Tree tree, Environment env) throws EvalException
   {
      int result = 0;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getSubTree(0), env );
      Value valueR = evaluateN( tree.getSubTree(1), env );

      if ( ! valueL.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueL );
      }
      if ( ! valueR.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueR );
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( node.equals("+") )
      {
         result = resultL + resultR;
      }
      else if ( node.equals("-") )
      {
         result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         result = resultL * resultR;
      }
      else if ( node.equals("/") )
      {
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         result = resultL % resultR;
      }
      else if ( node.equals("^") )
      {
         result = (int)Math.pow(resultL, resultR);
      }
      else
      {
         throw new EvalException("invalid arithmetic operator: " + node);
      }

      return new Value( "int", result );
   }//evaluateA()


   // Evaluate a boolean expression
   private static Value evaluateB(Tree tree, Environment env) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getSubTree(0), env );
      if ( ! valueL.tag.equals("bool") )
      {
         throw new EvalException("illegal boolean value: " + valueL );
      }
      boolean resultL = valueL.valueB;

      boolean resultR = false;
      if ( tree.degree() == 2 )  // check for second operand
      {
         Value valueR = evaluateN( tree.getSubTree(1), env );
         if ( ! valueR.tag.equals("bool") )
         {
            throw new EvalException("illegal boolean value: " + valueR );
         }
         resultR = valueR.valueB;
      }

      if ( node.equals("&&") )  // not short circuited
      {
         result = resultL && resultR;
      }
      else if ( node.equals("||") )  // not short circuited
      {
         result = resultL || resultR;
      }
      else if ( node.equals("!") )  // unary operator
      {
         result = ! resultL;
      }
      else
      {
         throw new EvalException("invalid boolean operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateB()


   // Evaluate an equality expression (which is a kind of boolean expression)
   private static Value evaluateEq(Tree tree, Environment env) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateE( tree.getSubTree(0), env );
      Value valueR = evaluateE( tree.getSubTree(1), env );

      if ( ! valueL.tag.equals(valueR.tag) )
      {
         result = false;
      }
      else if ( valueL.tag.equals("int") )
      {
         result = (valueL.valueI == valueR.valueI);
      }
      else if ( valueL.tag.equals("bool") )
      {
         result = (valueL.valueB == valueR.valueB);
      }
      else
      {
         throw new EvalException("invalid Value tag: " + valueL.tag);
      }

      if ( node.equals("!=") )
      {
         result = ! result;
      }
      else  if ( ! node.equals("==") )
      {
        throw new EvalException("invalid equality operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateEq()


   // Evaluate a relational expression (which is a kind of boolean expression)
   private static Value evaluateR(Tree tree, Environment env) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getSubTree(0), env );
      Value valueR = evaluateN( tree.getSubTree(1), env );

      if ( ! valueL.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueL );
      }
      if ( ! valueR.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueR );
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( node.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( node.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( node.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( node.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else
      {
         throw new EvalException("invalid relational operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateR()

}//Evaluate