/*
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= '(' 'prog' Define* Exp ')'

     Define ::= '(' 'define' VARIABLE NExp ')'
              | '(' 'define' VARIABLE Lambda ')'

     Lambda ::= '(' 'lambda' VARIABLE Exp ')'

        Exp ::= Begin | Var | NExp

      Begin ::= '(' 'begin' Exp  Exp+ ')'

        Var ::= '(' 'var' VARIABLE NExp ')'
              | '(' 'var' VARIABLE Lambda ')'

       NExp ::= Apply | Set | Print | AExp | BExp
              | VARIABLE | INTEGER | BOOLEAN

      Apply ::= '(' 'apply' VARIABLE NExp ')'

        Set ::= '(' 'set' VARIABLE NExp ')'

      Print ::= '(' 'print' NExp ')'

       AExp ::= '(' BinOp NExp NExp ')'

     BinOp  ::= '+' | '-' | '*' | '/' | '%' | '^'

       BExp ::= '(' BoolOp NExp NExp+ ')'
              | '(' '!' NExp ')'
              | '(' RelOp NExp NExp ')'
              | '('  EqOp NExp NExp ')'

     BoolOp ::= '&&' | '||'
     RelOp  ::= '<' | '>' | '<=' | '>='
      EqOp  ::= '==' | '!='

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*


   In order to allow nested functions, this language adds one new
   clause to the Var term from Language-7
  .
   Since this language now allows "local" functions, we need to have
   nested namespaces for function names (in addtion to the nested
   namespaces we already have for integer and bolean variables). We
   implement this by adding a nesting link to the FunctionEnvironment class

   The only changes to the implementation of Evaluate from Language-7 that we
   need to make are:
   1) Both evaluateBegin() and evluateApply() must now create a nested
      FunctionEnvironment object (in addition to the nested Environment object).

   2) All of the evaluate functions now take two environment paramters, one
      parameter for a Environment object and one for a FunctionEnvironment object.

   3) The definition of evaluateVar() must now implement the new clause in a Var term.
*/

public class Evaluate
{
   // Global environment for variable definitions.
   private static Environment globalEnv;
   // Global environment for function definitions.
   private static FunctionEnvironment funcEnv;

   public static Value evaluate(Tree tree) throws EvalException
   {
      globalEnv = new Environment();
      funcEnv = new FunctionEnvironment();

      return  evaluateProg( tree );
   }//evaluate()


   private static Value evaluateProg(Tree tree) throws EvalException
   {
      Value result = null;

      // Evaluate the definitions in the Prog term.
      // Each definition will have the side effect of putting
      // something in one of the global environments.
      for (int i = 0; i < tree.degree()-1; i++)
      {
         evaluateDefine( tree.getSubTree(i), globalEnv, funcEnv );
      }

      // Evaluate the single expression.
      result = evaluateE( tree.getSubTree(tree.degree()-1), globalEnv, funcEnv );

      return result;
   }//evaluateProg()


   // Evaluate definitions. Notice that there is no return value.
   // This function only mutates the global environment objects.
   private static void evaluateDefine(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      // get the variable or function name
      String name = tree.getSubTree(0).getElement();

      // get the expression
      Tree expr = tree.getSubTree(1);

      // determine if the definition is a function
      if (expr.getElement().equals("lambda"))
      {
         funcEnv.add(name, expr);
         System.out.println( funcEnv + "\n" ); // for debugging purposes
      }
      else // not a function, so evaluate the expression
      {
         Value value = evaluateN(expr, env, fenv);

         // check if this variable has already been defined
         if ( ! env.definedLocal(name) )
         {
            env.add(name, value);
         }
         else // this variable is already in the local Environment object
         {
            env.update(name, value);
         }
         System.out.println( env + "\n" ); // for debugging purposes
      }
   }//evaluateDefine()


   /**
      Notice that we evaluate an expression using the included environment which
      is a linked list of environment objects, terminating with the global environment.
   */
   private static Value evaluateE(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( node.equals("begin") )
      {
         result = evaluateBegin(tree, env, fenv);
      }
      else if ( node.equals("var") )
      {
         result = evaluateVar(tree, env, fenv);
      }
      else
      {
         result = evaluateN(tree, env, fenv);
      }

      return result;
   }//evaluateE()


   private static Value evaluateBegin(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      // create a new environment "nested" in the previous environment
      Environment localEnv = new Environment(env);
      FunctionEnvironment localFenv = new FunctionEnvironment(fenv);

      Value result = null;

      // evaluate each expression (using the nested environment)
      for (int i = 0; i < tree.degree(); i++)
      {
         // Notice that, except for the last expression, this evaluation
         // serves no purpose unless the expression is a "var" or "set" term,
         // which mutate environment objects, or a "begin" term which
         // contains a "var" or "set" term. In other words, unless this
         // expression has a side-effect, it does not do anything.
         result = evaluateE( tree.getSubTree(i), localEnv, localFenv );
      }

      // Notice that when this evaluateBegin() method returns, the
      // localEnv object becomes a garbage object, so it is implicitly
      // removed from the list of Environment objects.

      return result;
   }//evaluateBegin()


   private static Value evaluateVar(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      Value result = new Value("bool", false);  // default return value

      // get the variable
      String variable = tree.getSubTree(0).getElement();

      // get the expression
      Tree expr = tree.getSubTree(1);

      if (expr.getElement().equals("lambda") )
      {
         // check if this variable has already been defined
         if ( ! fenv.definedLocal(variable) )
         {
            fenv.add(variable, expr);
         }
         else // this variable is already in the local Environment object
         {
            throw new EvalException("local function already exisits: " + variable);
         }
         System.out.println( fenv + "\n" ); // for debugging purposes
      }
      else // not a function, so evaluate the expression
      {
         result = evaluateN(expr, env, fenv);

         // check if this variable has already been defined
         if ( !env.definedLocal(variable) )
         {
            env.add(variable, result);
         }
         else // this variable is already in the local Environment object
         {
            throw new EvalException("local variable already exisits: " + variable);
         }
         System.out.println( env + "\n" ); // for debugging purposes
      }

      return result;

   }//evaluateVar()


   private static Value evaluateN(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( "bool", node.equals("true") );
         }
         else if ( node.matches("^[0-9][0-9]*") )
         {
            result = new Value( "int", Integer.parseInt( node ) );
         }
         else if ( env.defined(node) )
         {
            result = env.lookUp( node );
         }
         else
         {
            throw new EvalException("undefined variable: " + node);
         }
      }
      else if ( node.equals("apply") )
      {
         result = evaluateApply(tree, env, fenv);
      }
      else if ( node.equals("set") )
      {
         result = evaluateSet(tree, env, fenv);
      }
      else if ( node.equals("print") )
      {
         result = evaluatePrint(tree, env, fenv);
      }
      else if ( node.equals("+") ||     // arithmetic operators
                node.equals("-") ||
                node.equals("*") ||
                node.equals("/") ||
                node.equals("%") ||
                node.equals("^") )
      {
         result = evaluateA(tree, env, fenv);
      }
      else if ( node.equals("&&") ||    // boolean operators
                node.equals("||") ||
                node.equals("!") )
      {
         result = evaluateB(tree, env, fenv);
      }
      else if ( node.equals("==") ||    // equality operators
                node.equals("!=") )
      {
         result = evaluateEq(tree, env, fenv);
      }
      else if ( node.equals("<") ||     // relational operators
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") )
      {
         result = evaluateR(tree, env, fenv);
      }
      else
      {
         throw new EvalException("invalid expression: " + AST2infix.ast2infix(tree));
      }

      return result;
   }//evaluateN()


   /**
      This method "applies" a function to an actual parameter.
      This method evaluates the body of the function in an environment
      that binds the actual parameter value (from this function application)
      to the formal parameter (from the function's definition).
   */
   private static Value evaluateApply(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      Value result;

      // Evaluate the second expression (using the original environments).
      Value actualParam = evaluateN( tree.getSubTree(1), env, fenv );

      // Retrive, from within the current environment,
      // a reference to the function's definition
      // (i.e., the function's "lambda expression").
      Tree function = fenv.lookUp( tree.getSubTree(0).getElement() );

      // Create a new environment "nested" in the original environment.
      Environment localEnv = new Environment(env);    // NOTE: dynamic scope
      FunctionEnvironment localFenv = new FunctionEnvironment(fenv);

      // Bind the actual parameter value to the formal parameter.
      // The binding is done in the new, local environment.
      String formalParam = function.getSubTree(0).getElement();
      localEnv.add(formalParam, actualParam);
      System.out.println( localEnv + "\n" ); // for debugging purposes

      // Finally, evaluate the body of the function using the
      // new environment (which contains the binding of the actual
      // parameter value to the function's formal parameter).
      result = evaluateE( function.getSubTree(1), localEnv, localFenv );

      return result;
   }//evaluateApply()


   private static Value evaluateSet(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      // get the variable
      String variable = tree.getSubTree(0).getElement();

      // get, and then evaluate, the expression
      Tree expr = tree.getSubTree(1);
      Value value = evaluateN(expr, env, fenv);

      // check if this variable has already been defined
      if ( !env.defined(variable) )
      {
         throw new EvalException("undefined variable: " + variable);
      }
      else // this variable is already in the environment
      {
         env.update(variable, value);
      }

      System.out.println( env + "\n" ); // for debugging purposes

      return value;

   }//evaluateSet()


   private static Value evaluatePrint(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      Value result = evaluateN( tree.getSubTree(0), env, fenv );

      System.out.println( result );

      return result;
   }//evaluatePrint()


   // Evaluate an arithmetic expression
   private static Value evaluateA(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      int result = 0;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getSubTree(0), env, fenv );
      Value valueR = evaluateN( tree.getSubTree(1), env, fenv );

      if ( ! valueL.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueL );
      }
      if ( ! valueR.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueR );
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( node.equals("+") )
      {
         result = resultL + resultR;
      }
      else if ( node.equals("-") )
      {
         result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         result = resultL * resultR;
      }
      else if ( node.equals("/") )
      {
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         result = resultL % resultR;
      }
      else if ( node.equals("^") )
      {
         result = (int)Math.pow(resultL, resultR);
      }
      else
      {
         throw new EvalException("invalid arithmetic operator: " + node);
      }

      return new Value( "int", result );
   }//evaluateA()


   // Evaluate a boolean expression
   private static Value evaluateB(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getSubTree(0), env, fenv );
      if ( ! valueL.tag.equals("bool") )
      {
         throw new EvalException("illegal boolean value: " + valueL );
      }
      boolean resultL = valueL.valueB;

      boolean resultR = false;
      if ( tree.degree() == 2 )  // check for second operand
      {
         Value valueR = evaluateN( tree.getSubTree(1), env, fenv );
         if ( ! valueR.tag.equals("bool") )
         {
            throw new EvalException("illegal boolean value: " + valueR );
         }
         resultR = valueR.valueB;
      }

      if ( node.equals("&&") )  // not short circuited
      {
         result = resultL && resultR;
      }
      else if ( node.equals("||") )  // not short circuited
      {
         result = resultL || resultR;
      }
      else if ( node.equals("!") )  // unary operator
      {
         result = ! resultL;
      }
      else
      {
         throw new EvalException("invalid boolean operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateB()


   // Evaluate an equality expression (which is a kind of boolean expression)
   private static Value evaluateEq(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateE( tree.getSubTree(0), env, fenv );
      Value valueR = evaluateE( tree.getSubTree(1), env, fenv );

      if ( ! valueL.tag.equals(valueR.tag) )
      {
         result = false;
      }
      else if ( valueL.tag.equals("int") )
      {
         result = (valueL.valueI == valueR.valueI);
      }
      else if ( valueL.tag.equals("bool") )
      {
         result = (valueL.valueB == valueR.valueB);
      }
      else
      {
         throw new EvalException("invalid Value tag: " + valueL.tag);
      }

      if ( node.equals("!=") )
      {
         result = ! result;
      }
      else  if ( ! node.equals("==") )
      {
        throw new EvalException("invalid equality operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateEq()


   // Evaluate a relational expression (which is a kind of boolean expression)
   private static Value evaluateR(Tree tree, Environment env, FunctionEnvironment fenv) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getSubTree(0), env, fenv );
      Value valueR = evaluateN( tree.getSubTree(1), env, fenv );

      if ( ! valueL.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueL );
      }
      if ( ! valueR.tag.equals("int") )
      {
         throw new EvalException("illegal arithmetic value: " + valueR );
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( node.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( node.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( node.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( node.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else
      {
         throw new EvalException("invalid relational operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateR()

}//Evaluate