/*
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

        Exp ::= AExp | BExp

       AExp ::= '(' BinOp AExp AExp ')'
              | INTEGER

      BinOp ::= '+' | '-' | '*' | '/' | '%'

       BExp ::= '(' BoolOp BExp BExp ')'
              | '(' '!' BExp '()' ')'      // notice how ugly this unary ! operator is
              | '(' RelOp AExp AExp ')'
              | '(' EqOp  Exp  Exp  ')'
              | BOOLEAN

     BoolOp ::= '&&' | '||'
      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'


   Notice that the result of evaluating an Exp is either a
   integer value or a boolean value.

   By using the language of Trees (instead of the language of BTrees)
   you can implement the following improvements in this interpreter.

   Extension 1: Create unary +, - operators (and fix !).

   Extension 2: Make +, *, &&, || into operators that take two or more parameters.

   Extension 3: Short circuit the evaluation of the boolean operators && and ||.
*/

public class Evaluate
{
   /**
      The methods evaluate(), evaluateA(), evaluateB(), and evaluateR() are
      essentially a post-order traversal of the binary tree.
   */
   public static Value evaluate(BTree tree)
   {
      Value result = null;

      String node = tree.getElement();

      if ( tree.degree() == 0 )  // constant term
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( "bool", node.equals("true") );
         }
         else if ( node.matches("^[0-9][0-9]*") )
         {
            result = new Value( "int", Integer.parseInt( node ) );
         }
         else
         {
            throw new Error("invalid expression: " + AST2infix.ast2infix(tree));
         }
      }
      else if ( node.equals("+") ||
                node.equals("-") ||
                node.equals("*") ||
                node.equals("/") ||
                node.equals("%") )
      {
         result = evaluateA(tree);  // arithmetic operator
      }
      else if ( node.equals("&&") ||
                node.equals("||") ||
                node.equals("!") )
      {
         result = evaluateB(tree);  // boolean operator
      }
      else if ( node.equals("==") ||
                node.equals("!=") )
      {
         result = evaluateEq(tree);  // equality operator
      }
      else if ( node.equals("<") ||
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") )
      {
         result = evaluateR(tree);  // relational operator
      }
      else
      {
         throw new Error("invalid operator: " + AST2infix.ast2infix(tree));
      }

      return result;
   }//evaluate()


   // Evaluate an arithmetic expression
   private static Value evaluateA(BTree tree)
   {
      int result = 0;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         result = Integer.parseInt( node );
      }
      else
      {
         int resultL = evaluateA( tree.getLeftTree() ).valueI;
         int resultR = evaluateA( tree.getRightTree() ).valueI;

         if ( node.equals("+") )
         {
            result = resultL + resultR;
         }
         else if ( node.equals("-") )
         {
            result = resultL - resultR;
         }
         else if ( node.equals("*") )
         {
            result = resultL * resultR;
         }
         else if ( node.equals("/") )
         {
            result = resultL / resultR;
         }
         else if ( node.equals("%") )
         {
            result = resultL % resultR;
         }
         else
         {
           throw new Error("invalid arithmetic operator: " + node);
         }
      }

      return new Value( "int", result );
   }//evaluateA()


   // Evaluate a boolean expression
   private static Value evaluateB(BTree tree)
   {
      boolean result = false;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         result = node.equals("true");
      }
      else if ( node.equals("==") ||  // check for equality operators
                node.equals("!=") )
      {
         result = evaluateEq( tree ).valueB;
      }
      else if ( node.equals("<") ||  // check for relational operators
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") )
      {
         result = evaluateR( tree ).valueB;
      }
      else  // boolean operator
      {
         boolean resultL = evaluateB( tree.getLeftTree() ).valueB;
         boolean resultR = false;

         if ( tree.degree() == 2 )
            resultR = evaluateB( tree.getRightTree() ).valueB;

         if ( node.equals("&&") )      // not short circuited
         {
            result = resultL && resultR;
         }
         else if ( node.equals("||") ) // not short circuited
         {
            result = resultL || resultR;
         }
         else if ( node.equals("!") )
         {
            result = ! resultL;
         }
         else
         {
           throw new Error("invalid boolean operator: " + node);
         }
      }

      return new Value( "bool", result );
   }//evaluateB()


   // Evaluate an equality expression (which is a kind of boolean expression)
   private static Value evaluateEq(BTree tree)
   {
      boolean result = false;

      String op = tree.getElement();

      Value valueL = evaluate( tree.getLeftTree() );
      Value valueR = evaluate( tree.getRightTree() );

      if ( ! valueL.tag.equals(valueR.tag) )
      {
         result = false;
      }
      else if ( valueL.tag.equals("int") )
      {
         result = (valueL.valueI == valueR.valueI);
      }
      else if ( valueL.tag.equals("bool") )
      {
         result = (valueL.valueB == valueR.valueB);
      }
      else
      {
         throw new Error("invalid Value tag: " + valueL.tag);
      }

      if ( op.equals("!=") )
      {
         result = ! result;
      }
      else  if ( ! op.equals("==") )
      {
        throw new Error("invalid equality operator: " + op);
      }

      return new Value( "bool", result );
   }//evaluateEq()


   // Evaluate a relational expression (which is a kind of boolean expression)
   private static Value evaluateR(BTree tree)
   {
      boolean result = false;

      String op = tree.getElement();

      int resultL = evaluateA( tree.getLeftTree() ).valueI;
      int resultR = evaluateA( tree.getRightTree() ).valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else
      {
        throw new Error("invalid relational operator: " + op);
      }

      return new Value( "bool", result );
   }//evaluateR()

}//Evaluate