/*
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

        Exp ::= Begin | Var | NExp

      Begin ::= '(' 'begin' Exp  Exp ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

       NExp ::= AExp | BExp | VARIABLE | INTEGER | BOOLEAN

       AExp ::= '(' BinOp  NExp NExp ')'

     BinOp  ::= '+' | '-' | '*' | '/' | '%'

       BExp ::= '(' BoolOp NExp NExp ')'
              | '(' '!' NExp '()' ')'      // notice how ugly this unary ! operator is
              | '(' RelOp NExp NExp ')'
              | '(' EqOp  NExp NExp ')'

     BoolOp ::= '&&' | '||'
      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*

   This language has exactly the same syntax as language Language-4,
   but the semantics of this language are different.
*/

public class Evaluate
{
   private static Environment globalEnv;  // global Environment object


   public static Value evaluate(BTree tree)
   {
      globalEnv = new Environment();

      return  evaluateE( tree, globalEnv ); // evaluate the expression using the global env
   }//evaluate()


   /**
      The methods evaluateE(), evaluateN(), evaluateA(), evaluateB(), and evaluateR() are
      essentially a post-order traversal of the binary tree.

      Notice that we valuate an expression using the included environment which
      is a linked list of environment objects, terminating with the global environment.
   */
   private static Value evaluateE(BTree tree, Environment env)
   {
      Value result = null;

      String node = tree.getElement();

      if ( node.equals("begin") )
      {
         result = evaluateBegin(tree, env);
      }
      else if ( node.equals("var") )
      {
         result = evaluateVar(tree, env);
      }
      else
      {
         result = evaluateN(tree, env);
      }

      return result;
   }//evaluateE()


   private static Value evaluateBegin(BTree tree, Environment env)
   {
      // create a new environment "nested" in the previous environment
      Environment localEnv = new Environment(env);

      Value result;

      // evaluate the first expression (using the nested environment)
      //
      // Notice that this evaluation serves no purpose unless
      // the first expression is a "var" term, which mutates the
      // local environment object, or a "begin" term which contains
      // a "var" term. In other words, unless this first expression
      // has a side-effect, it does not do anything.
      result = evaluateE( tree.getLeftTree(), localEnv );

      // evaluate the second expression (using the nested environment)
      result =  evaluateE( tree.getRightTree(), localEnv );

      // Notice that when this evaluateBegin() method returns, the
      // localEnv object becomes a garbage object, so it is implicitly
      // removed from the list of Environment objects.

      return result;
   }//evaluateBegin()


   private static Value evaluateVar(BTree tree, Environment env)
   {
      Value result = null;

      // get the variable
      String variable = tree.getLeftTree().getElement();

      // get, and then evaluate, the expression
      BTree expr = tree.getRightTree();
      result = evaluateE(expr, env);

      // check if this variable has already been defined
      if ( ! env.definedLocal(variable) )
      {
         env.add(variable, result);
      }
      else // this variable is already in the local Environment object
      {
         env.update(variable, result);
      }

      System.out.println( env + "\n" ); // for debugging purposes

      return result;

   }//evaluateVar()


   private static Value evaluateN(BTree tree, Environment env)
   {
      Value result = null;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( "bool", node.equals("true") );
         }
         else if ( node.matches("^[0-9][0-9]*") )
         {
            result = new Value( "int", Integer.parseInt( node ) );
         }
         else if ( env.defined(node) )
         {
            result = env.lookUp( node );
         }
         else
         {
            throw new Error("invalid expression: " + AST2infix.ast2infix(tree));
         }
      }
      else if ( node.equals("+") ||     // arithmetic operators
                node.equals("-") ||
                node.equals("*") ||
                node.equals("/") ||
                node.equals("%") )
      {
         result = evaluateA(tree, env);
      }
      else if ( node.equals("&&") ||    // boolean operators
                node.equals("||") ||
                node.equals("!") )
      {
         result = evaluateB(tree, env);
      }
      else if ( node.equals("==") ||    // equality operators
                node.equals("!=") )
      {
         result = evaluateEq(tree, env);
      }
      else if ( node.equals("<") ||     // relational operators
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") )
      {
         result = evaluateR(tree, env);
      }
      else
      {
         throw new Error("invalid expression: " + AST2infix.ast2infix(tree));
      }

      return result;
   }//evaluateN()


   // Evaluate an arithmetic expression
   private static Value evaluateA(BTree tree, Environment env)
   {
      int result = 0;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getLeftTree(), env );
      Value valueR = evaluateN( tree.getRightTree(), env );

      if ( ! valueL.tag.equals("int") )
      {
         throw new Error ("illegal arithmetic value: " + valueL );
      }
      if ( ! valueR.tag.equals("int") )
      {
         throw new Error ("illegal arithmetic value: " + valueR );
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( node.equals("+") )
      {
         result = resultL + resultR;
      }
      else if ( node.equals("-") )
      {
         result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         result = resultL * resultR;
      }
      else if ( node.equals("/") )
      {
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         result = resultL % resultR;
      }
      else
      {
         throw new Error("invalid arithmetic operator: " + node);
      }

      return new Value( "int", result );
   }//evaluateA()


   // Evaluate a boolean expression
   private static Value evaluateB(BTree tree, Environment env)
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getLeftTree(), env );
      if ( ! valueL.tag.equals("bool") )
      {
         throw new Error ("illegal boolean value: " + valueL );
      }
      boolean resultL = valueL.valueB;

      boolean resultR = false;
      if ( tree.degree() == 2 )  // check for second operand
      {
         Value valueR = evaluateN( tree.getRightTree(), env );
         if ( ! valueR.tag.equals("bool") )
         {
            throw new Error ("illegal boolean value: " + valueR );
         }
         resultR = valueR.valueB;
      }

      if ( node.equals("&&") )  // not short circuited
      {
         result = resultL && resultR;
      }
      else if ( node.equals("||") )  // not short circuited
      {
         result = resultL || resultR;
      }
      else if ( node.equals("!") )  // unary operator
      {
         result = ! resultL;
      }
      else
      {
         throw new Error("invalid boolean operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateB()


   // Evaluate an equality expression (which is a kind of boolean expression)
   private static Value evaluateEq(BTree tree, Environment env)
   {
      boolean result = false;

      String op = tree.getElement();

      Value valueL = evaluateN( tree.getLeftTree(), env );
      Value valueR = evaluateN( tree.getRightTree(), env );

      if ( ! valueL.tag.equals(valueR.tag) )
      {
         result = false;
      }
      else if ( valueL.tag.equals("int") )
      {
         result = (valueL.valueI == valueR.valueI);
      }
      else if ( valueL.tag.equals("bool") )
      {
         result = (valueL.valueB == valueR.valueB);
      }
      else
      {
         throw new Error("invalid Value tag: " + valueL.tag);
      }

      if ( op.equals("!=") )
      {
         result = ! result;
      }
      else  if ( ! op.equals("==") )
      {
        throw new Error("invalid equality operator: " + op);
      }

      return new Value( "bool", result );
   }//evaluateEq()


   // Evaluate a relational expression (which is a kind of boolean expression)
   private static Value evaluateR(BTree tree, Environment env)
   {
      boolean result = false;

      String node = tree.getElement();

      Value valueL = evaluateN( tree.getLeftTree(), env );
      Value valueR = evaluateN( tree.getRightTree(), env );

      if ( ! valueL.tag.equals("int") )
      {
         throw new Error ("illegal arithmetic value: " + valueL );
      }
      if ( ! valueR.tag.equals("int") )
      {
         throw new Error ("illegal arithmetic value: " + valueR );
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( node.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( node.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( node.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( node.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else if ( node.equals("==") )
      {
         result = resultL == resultR;
      }
      else if ( node.equals("!=") )
      {
         result = resultL != resultR;
      }
      else
      {
         throw new Error("invalid relational operator: " + node);
      }

      return new Value( "bool", result );
   }//evaluateR()

}//Evaluate