/**
   A FunctionEnvironment object holds <variable,tree> pairs and
   has a link to its nesting FunctionEnvironment object (its outer scope).

   The "tree" that we store in the <variable.tree> pair is a
   reference to the abstract syntax tree of the Lambda term
   that defines the function.
*/
import java.util.ArrayList;

public class FunctionEnvironment
{
   private ArrayList<String> variables;
   private ArrayList<Tree>   trees;
   private FunctionEnvironment nestingLink;


   // Constructors
   public FunctionEnvironment()
   {
      variables = new ArrayList<String>();
      trees     = new ArrayList<Tree>();
      nestingLink = null;
   }

   public FunctionEnvironment(FunctionEnvironment env)
   {
      variables = new ArrayList<String>();
      trees     = new ArrayList<Tree>();
      nestingLink = env;
   }

   /**
      Add a <variable, tree> pair to this environment object.
   */
   public void add(String variable, Tree tree)
   {
      variables.add(variable);
          trees.add(tree);
   }

   /**
      Look up variable in the environment chain and
      return its associated Tree.

      Throws an Error if variable is not found.
   */
   public Tree lookUp(String variable)
   {
      int i;
      for (i=0; i < variables.size(); i++)
         if (variable.equals(variables.get(i)))
            break;

      if ( i < variables.size() )
      {
         return trees.get(i);
      }
      else
      {
         if (nestingLink == null)
            throw new Error("undefined variable: " + variable);
         else
            return nestingLink.lookUp(variable);
      }
   }

   /**
      Look up variable in the environment chain.
      Return true if the variable is in the chain,
      otherwise return false.
   */
   public boolean defined(String variable)
   {
      int i;
      for (i=0; i < variables.size(); i++)
         if (variable.equals(variables.get(i)))
            break;

      if ( i < variables.size() )
      {
         return true;
      }
      else
      {
         if (nestingLink == null)
            return false;
         else
            return nestingLink.defined(variable);
      }
   }


   /**
      Look up variable in this environment object.
      Return true if the variable is in this object,
      otherwise return false.
   */
   public boolean definedLocal(String variable)
   {
      int i;
      for (i=0; i < variables.size(); i++)
         if (variable.equals(variables.get(i)))
            break;

      if ( i < variables.size() )
      {
         return true;
      }
      else
      {
         return false;
      }
   }


   /**
      Convert the contents of the environment chain into a string.
      This is mainly for debugging purposes.
   */
   public String toString()
   {
      String result = "";

      if (nestingLink == null)
      {
         result += "[Global Function Environment";
      }
      else
      {
         result += "[Nested Function Environment";
      }

      for (int i = 0; i < variables.size(); i++)
      {
         result += "\n " + variables.get(i) + " = " + trees.get(i);
      }
      result += "]";

      if (nestingLink == null)
      {
         return result;
      }
      else
      {
         return result + "\n||\n\\/\n" + nestingLink.toString();
      }
   }

}//FunctionEnvironment