/*
   This program is a static, semantic analyzer that does type checking for
   the following language. (This is Language-14t after applying the prepocessor.)

       Prog ::= '(' 'prog' Body ')'

       Body ::= Exp+

        Exp ::= Begin | Var | NExp

      Begin ::= '(' 'begin' Body ')'

        Var ::= '(' 'var' Type VARIABLE NExp ')'     // variables have a "type"

       NExp ::= Set | If | While | And | Or | Print
              | Apply | VARIABLE
              | Lambda | INTEGER | BOOLEAN

        Set ::= '(' 'set' VARIABLE NExp ')'

         If ::= '(' 'if' NExp Exp Exp ')'

      While ::= '(' 'while' NExp Exp ')'

        And ::= '(' '&&' NExp NExp NExp* ')'

         Or ::= '(' '||' NExp NExp NExp* ')'

         Eq ::= '(' '==' NExp NExp ')'

      NotEq ::= '(' '!=' NExp NExp ')'

      Print ::= '(' 'print' NExp ')'

      Apply ::= '(' 'apply' NExp NExp* ')'

     Lambda ::= '(' 'lambda' Signature VARIABLE* Exp ')'  // functions have a "signature"

  Signature ::= '(' 'signature' Type Type* ')'   // return type and parameter types

       Type ::= 'bool' | 'int' | Signature

   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
*/

public class TypeChecker
{
   // Global environment for function and variable definitions.
   private static Environment globalEnv;

   // Global flag that records presence of type errors
   private static boolean typeChecked = true;


   public static boolean typeCheck(Tree tree)
   {
      typeChecked = true;

      globalEnv = new Environment();

      // Populate the global environment with the built-in functions.
      globalEnv.add("+", new Value("builtin", "+") );
      globalEnv.add("-", new Value("builtin", "-") );
      globalEnv.add("neg", new Value("builtin", "neg") );    // notice the new name
      globalEnv.add("*", new Value("builtin", "*") );
      globalEnv.add("/", new Value("builtin", "/") );
      globalEnv.add("%", new Value("builtin", "%") );
      globalEnv.add("^", new Value("builtin", "^") );
      globalEnv.add("!", new Value("builtin", "!") );
      globalEnv.add("<", new Value("builtin", "<") );
      globalEnv.add(">", new Value("builtin", ">") );
      globalEnv.add("<=", new Value("builtin", "<=") );
      globalEnv.add(">=", new Value("builtin", ">=") );
      //globalEnv.add("==", new Value("builtin", "==") );       // make this a special form
      //globalEnv.add("!=", new Value("builtin", "!=") );       // make this a special form
      //globalEnv.add("print", new Value("builtin", "print") ); // make this a special form

      checkProg( tree, globalEnv );

      return  typeChecked;
   }//typeCheck()


   private static Tree checkProg(Tree tree, Environment env)
   {
      Tree result = null;

      // Check the expressions in the Prog term.
      for (int i = 0; i < tree.degree(); i++)
      {
         result = checkE( tree.getSubTree(i), env );
      }

      return result;
   }//evaluateProg()


   private static Tree checkE(Tree tree, Environment env)
   {
      Tree result = null;

      if ( tree.getElement().equals("begin") )
      {
         result = checkBegin(tree, env);
      }
      else if ( tree.getElement().equals("var") )
      {
         result = checkVar(tree, env);
      }
      else
      {
         result = checkN(tree, env);
      }

      return result;
   }//checkE()


   private static Tree checkBegin(Tree tree, Environment env)
   {
      // create a new environment "nested" in the previous environment
      Environment localEnv = new Environment(env);

      Tree result = null;

      // check each expression (using the nested environment)
      for (int i = 0; i < tree.degree(); i++)
      {
         result = checkE( tree.getSubTree(i), localEnv );
      }

      return result;
   }//checkBegin()


   private static Tree checkVar(Tree tree, Environment env)
   {
      // get the variable's type
      Tree varType = tree.getSubTree(0);

      // get the variable name
      String name = tree.getSubTree(1).getElement();

      // get the type of the expression
      Tree exprType = checkN( tree.getSubTree(2), env );

      if ( ! compareTypes(varType, exprType) )
      {
         System.out.println("Type error: types don't agree: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      // check if this variable has already been defined
      if ( ! env.definedLocal(name) )
      {
         if ( varType.getElement().equals("bool") )
         {
            env.add(name, new Value("bool", false) );
         }
         else if ( varType.getElement().equals("int") )
         {
            env.add(name, new Value("int", 0) );
         }
         else if ( varType.getElement().equals("signature") )
         {
            Tree dummyLambda = new Tree("lambda");
            dummyLambda.addSubTree(varType);
            env.add( name, new Value("closure", new IPEP(dummyLambda, null)));
         }
      }
      else // this variable is already in the local Environment object
      {
         System.out.println("Syntax error: local variable already exisits: " + name);
         typeChecked = false;
      }
      System.out.println( env + "\n" ); // for debugging purposes

      return varType;
   }//checkVar()


   private static Tree checkN(Tree tree, Environment env)
   {
      Tree resultType = null;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
         {
            resultType = new Tree("bool");
         }
         else if ( node.matches("^[0-9][0-9]*") )
         {
            resultType = new Tree("int");
         }
         else if ( env.defined(node) )
         {
            // look up this variable's (dummy) value
            Value value = env.lookUp(node);

            if ( value.tag.equals("int") || value.tag.equals("bool") )
            {
               resultType = new Tree(value.tag);
            }
            else if ( value.tag.equals("closure") ) // need to return a signature
            {
               // Retrive, from within the value object,
               // a reference to the function's IPEP pair.
               IPEP ipepPair = value.valueC;

               // Retrive, from within the IPEP pair,
               // a reference to the function's definition
               // (i.e., the function's "lambda expression").
               Tree lambdaExp = ipepPair.exp;

               // Retrive, from within the function's lambda expression,
               // the function's signature.
               resultType = lambdaExp.getSubTree(0);
            }
            else if ( value.tag.equals("builtin") ) // need to return a signature
            {
               if ( value.valueBi.equals("+")
                 || value.valueBi.equals("-")
                 || value.valueBi.equals("*")
                 || value.valueBi.equals("/")
                 || value.valueBi.equals("%")
                 || value.valueBi.equals("^") )
               {
                  // create a signature for these operators
                  resultType = new Tree("signature");
                  resultType.addSubTree( new Tree("int") );
                  resultType.addSubTree( new Tree("int") );
                  resultType.addSubTree( new Tree("int") );
               }
               else if ( value.valueBi.equals("neg") )
               {
                  // create a signature for these operators
                  resultType = new Tree("signature");
                  resultType.addSubTree( new Tree("int") );
                  resultType.addSubTree( new Tree("int") );
               }
               else if ( value.valueBi.equals("!") )
               {
                  // create a signature for these operators
                  resultType = new Tree("signature");
                  resultType.addSubTree( new Tree("bool") );
                  resultType.addSubTree( new Tree("bool") );
               }
               else if ( value.valueBi.equals("<")
                      || value.valueBi.equals(">")
                      || value.valueBi.equals("<=")
                      || value.valueBi.equals(">=") )
               {
                  // create a signature for these operators
                  resultType = new Tree("signature");
                  resultType.addSubTree( new Tree("bool") );
                  resultType.addSubTree( new Tree("int") );
                  resultType.addSubTree( new Tree("int") );
               }
               else  // should never get here
               {
                  System.out.println("Type error: illegal built-in operator: " + node);
                  // return a default signature
                  resultType = new Tree("signature");
                  resultType.addSubTree( new Tree("int") );
                  resultType.addSubTree( new Tree("int") );
                  resultType.addSubTree( new Tree("int") );
                  typeChecked = false;
               }
            }
         }
         else  // undefined variable
         {
            System.out.println("Syntax error: undefined variable: " + node);
            resultType = new Tree("int");  // a default type (this isn't really a good idea?)
            typeChecked = false;
         }
      }
      else if ( node.equals("set") )
      {
         resultType = checkSet( tree, env );
      }
      else if ( node.equals("if") )
      {
         resultType = checkIf( tree, env );
      }
      else if ( node.equals("while") )
      {
         resultType = checkWhile( tree, env );
      }
      else if ( node.equals("&&") || node.equals("||") )
      {
         resultType = checkAndOr( tree, env );
      }
      else if ( node.equals("==") || node.equals("!=") )
      {
         resultType = checkEqNotEq( tree, env );
      }
      else if ( node.equals("print") )
      {
         resultType = checkPrint( tree, env );
      }
      else if ( node.equals("apply") )
      {
         resultType = checkApply( tree, env) ;
      }
      else if ( node.equals("lambda") )
      {
         resultType = checkLambda( tree, env );
      }
      else
      {
         System.out.println("Syntax error: invalid expression: " + AST2infix.ast2infix(tree));
         typeChecked = false;
      }

      return resultType;
   }//checkN()


   private static Tree checkSet(Tree tree, Environment env)
   {
      // get the variable name
      String variable = tree.getSubTree(0).getElement();

      // get the type of the expression
      Tree exprType = checkE(tree.getSubTree(1), env);

      // check if this variable has been defined
      if ( ! env.defined(variable) )
      {
         System.out.println("Syntax error: undefined variable: " + variable);
         typeChecked = false;
      }
      else // this variable is in the environment
      {
         // get this variable's type
         Tree varType = checkN( tree.getSubTree(0), env );

         if ( ! compareTypes(varType, exprType) )
         {
            System.out.println("Type error: " + AST2infix.ast2infix(tree) );
            typeChecked = false;
         }
      }

      return exprType;
   }//checkSet()


   private static Tree checkIf(Tree tree, Environment env)
   {
      Tree condType = checkN( tree.getSubTree(0), env );

      // Check that the conditional expression has boolean type
      if ( ! condType.getElement().equals("bool") )
      {
         System.out.println("Type error: conditional must be boolean: "
                              + AST2infix.ast2infix(tree.getSubTree(0)) );
         typeChecked = false;
      }

      Tree expr1Type = checkN( tree.getSubTree(1), env );
      Tree expr2Type = checkN( tree.getSubTree(2), env );

      // Check that the two expressions have the same type
      if ( ! compareTypes( expr1Type, expr2Type ) )
      {
         System.out.println("Type error: if-expression not well typed: "
                              + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      return expr1Type;
   }//checkIf()


   private static Tree checkWhile(Tree tree, Environment env)
   {
      Tree condType = checkN( tree.getSubTree(0), env );

      // Check that the conditional expression has boolean type
      if ( ! condType.getElement().equals("bool") )
      {
         System.out.println("Type error: conditional must be boolean: "
                              + AST2infix.ast2infix(tree.getSubTree(0)) );
         typeChecked = false;
      }

      // Check that the expression is well typed.
      // We do not need to do anything with its type.
      checkN( tree.getSubTree(1), env );

      return new Tree("bool"); // always return a boolean type
   }//checkWhile()


   private static Tree checkAndOr(Tree tree, Environment env)
   {
      // Check that every operand has type boolean.
      for (int i = 0; i < tree.degree(); i++)
      {
         Tree exprType = checkN( tree.getSubTree(i), env );
         if ( ! exprType.getElement().equals("bool") )
         {
            System.out.println("Type error: operands to && must be boolean: "
                                 + AST2infix.ast2infix(tree.getSubTree(i)) );
            typeChecked = false;
         }
      }

      return new Tree("bool"); // always return a boolean type
   }//checkAndOr()


   // NOTE: This is a special form not because of the need for
   // a special evaluation rule, but because it needs a special
   // type checking rule (i.e., == and != don't have valid signatures).
   private static Tree checkEqNotEq(Tree tree, Environment env)
   {
      // All we need to do is make sure that
      // each parameter is well typed.
      // We do not need to compare those types.
      checkN( tree.getSubTree(0), env );
      checkN( tree.getSubTree(1), env );

      return new Tree("bool"); // always return a boolean type
   }//checkEqNotEq()


   // NOTE: This is a special form not because of the need for
   // a special evaluation rule, but because it needs a special
   // type checking rule (i.e., print does not have a valid signature).
   private static Tree checkPrint(Tree tree, Environment env)
   {
      return checkE( tree.getSubTree(0), env );
   }//checkPrint()


   /**
      This method checks the application of a function to actual parameters.
      This method checks that the declared types of the formal parameters match
      the type the actual parameters in this application. The type returned by
      this method is the function's declared return type, found in the function's
      Signature.
   */
   private static Tree checkApply(Tree tree, Environment env)
   {
      Tree returnType = null;

      // Get the function's signature.
      // Notice that a function signature can come from one of three kinds of expressions.
      // 1) A variable that evaluates to a function.
      // 2) A function application that returns a function value.
      // 3) An anonymous lambda application.
      Tree signature = checkN( tree.getSubTree(0), env );

      // Check that that we really have a function definition.
      if ( ! signature.getElement().equals("signature") )
      {
         System.out.println("Type error: not a function: "
                             + AST2infix.ast2infix(tree) );
         // Choose an arbitrary return type.
         returnType = new Tree("int");
         typeChecked = false;
      }
      else  // we really do have a function value
      {
         // Check that there are as many actual parameters
         // as the number of formal parameters.
         if ( signature.degree() != tree.degree() )
         {
            System.out.println("Syntax error: wrong number of parameters: "
                                + AST2infix.ast2infix(tree) );
            typeChecked = false;
         }

         // Check that the actual parameter types agree with the formal parameter types.
         for (int i = 1; i < signature.degree(); i++)
         {
            // Get the type of an actual parameter.
            Tree actualType = checkN( tree.getSubTree(i), env );
            // Compare the actual parameter type with the formal parameter type.
            if ( ! compareTypes( actualType, signature.getSubTree(i) ) )
            {
               System.out.println("Type error: wrong actual parameter type: "
                                   + AST2infix.ast2infix(tree) );
               typeChecked = false;
            }
         }

         // Get the return type.
         returnType = signature.getSubTree(0);
      }

      return returnType;
   }//checkApply()


   /**
      Type check a lambda expression. Make sure the type of the body
      of the function agrees with the type of the return value found
      in the Signature.
      The return value is the lambda expression's signature.
   */
   private static Tree checkLambda(Tree tree, Environment env)
   {
      // Retrive, from within the function's lambda expression,
      // the function's signature.
      Tree signature = tree.getSubTree(0);

      // Retrive, from within the function's signature,
      // the function's return type.
      Tree returnType = signature.getSubTree(0);

      // Create a local environment with the formal parameters bound
      // to dummy actual parameters (that each have the correct type).
      Environment localEnv = new Environment(env);  // NOTE: static scope

      for (int i = 1; i < signature.degree(); i++)
      {
         // Retrive, from within the function's signature,
         // a formal parameter type.
         Tree formalType = signature.getSubTree(i);

         // Retrive, from within the function's lambda expression,
         // a formal parameter.
         String formalParam = tree.getSubTree(i).getElement();

         // Create a "dummy" actual parameter of the correct type.
         Value actualParam = null;
         if ( formalType.getElement().equals("bool") )
         {
            actualParam = new Value("bool", false);
         }
         else if ( formalType.getElement().equals("int") )
         {
            actualParam = new Value("int", 0);
         }
         else if ( formalType.getElement().equals("signature") )
         {
            Tree dummyLambda = new Tree("lambda");
            dummyLambda.addSubTree(formalType);
            actualParam = new Value("closure", new IPEP(dummyLambda, null));
         }
         // put the <parameter,value> pair in the local environment
         localEnv.add(formalParam, actualParam );
      }
      System.out.println( localEnv + "\n" ); // for debugging purposes

      // Retrive, from within the function's lambda expression,
      // the body of the function.
      Tree expr = tree.getSubTree(tree.degree()-1);

      // Check the type of the function's body (using the local
      // environment with the dummy bindings).
      Tree exprType = checkE(expr, localEnv);

      // Check that the type of the function's body
      // agrees with the function's return type.
      if ( ! compareTypes(exprType, returnType) )
      {
         System.out.println("Type error: wrong return type in lambda expression: "
                             + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      return signature;
   }//checkLambda()


   // Check that two types are exactly the same.
   // Do a (simultaneous) pre-order traversal of the type trees.
   private static boolean compareTypes(Tree tree1, Tree tree2)
   {
      boolean result = true;
      if ( ! tree1.getElement().equals( tree2.getElement() ) )
      {
         return false;
      }
      else if ( tree1.degree() != tree2.degree() )
      {
         return false;
      }
      else
      {
         for(int i = 0; result && (i < tree1.degree()); i++)
         {
            result = compareTypes( tree1.getSubTree(i), tree2.getSubTree(i) );
         }
      }
      return result;
   }//compareTypes()

}//TypeChecker