/*
   This program is a static, semantic analyzer that does type checking for
   the following language. (This is Language-10t after applying the prepocessor.)

       Prog ::= '(' 'prog' Define* Exp ')'

     Define ::= '(' 'define' Type VARIABLE NExp ')'       // variables have a "type"
              | '(' 'define' 'func' FUNC-NAME Lambda ')'  // functions have a "signature"

     Lambda ::= '(' 'lambda' Signature VARIABLE Exp ')'

  Signature ::= '(' 'signature' Type Type ')'  // return type followed by parameter type

       Type ::= 'bool' | 'int'

        Exp ::= Begin | Var | NExp

      Begin ::= '(' 'begin' Exp  Exp+ ')'

        Var ::= '(' 'var' Type VARIABLE NExp ')'       // variables have a "type"
              | '(' 'var' 'func' FUNC-NAME Lambda ')'  // functions have a "signature"

       NExp ::= Apply | Set | Print | AExp | BExp
              | VARIABLE | INTEGER | BOOLEAN

      Apply ::= '(' 'apply' FUNC-NAME NExp ')'

        Set ::= '(' 'set' VARIABLE NExp ')'

      Print ::= '(' 'print' NExp ')'

       AExp ::= '(' BinOp NExp NExp ')'

     BinOp  ::= '+' | '-' | '*' | '/' | '%' | '^'

       BExp ::= '(' BoolOp NExp NExp+ ')'
              | '(' '!' NExp ')'
              | '(' RelOp NExp NExp ')'
              | '('  EqOp NExp NExp ')'

     BoolOp ::= '&&' | '||'
     RelOp  ::= '<' | '>' | '<=' | '>='
      EqOp  ::= '==' | '!='

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
  FUNC-NAME ::= [a-zA-Z][a-zA-Z0-9]*

*/

public class TypeChecker
{
   // Global environment for function and variable definitions.
   private static Environment globalEnv;
   // Global flag that records presence of type errors
   private static boolean typeChecked = true;


   public static boolean typeCheck(Tree tree)
   {
      typeChecked = true;

      globalEnv = new Environment();

      checkProg( tree );

      return  typeChecked;
   }//typeCheck()


   private static String checkProg(Tree tree)
   {
      String result = null;

      // Evaluate the definitions in the Prog term.
      // Each definition will have the side effect of putting
      // something in one of the global environments.
      for (int i = 0; i < tree.degree()-1; i++)
      {
         checkDefine( tree.getSubTree(i), globalEnv );
      }

      // Evaluate the single expression.
      result = checkE( tree.getSubTree(tree.degree()-1), globalEnv );

      return result;
   }//evaluateProg()


   /**
      Check definitions. Notice that there is no return value.
      Definitions are not expressions, so they don;t have a type.
   */
   private static void checkDefine(Tree tree, Environment env)
   {
      // get the type declaration
      String declrType = tree.getSubTree(0).getElement();

      // determine if the definition is a function
      if (declrType.equals("func"))
      {
         // get the function name
         String name = tree.getSubTree(1).getElement();

         // check that the lambda expression is type correct
         if ( ! checkLambda( tree.getSubTree(2), env ) )
         {
            System.out.println("Type error: incorrect return type in " + name);
            typeChecked = false;
         }

         // add the function to the environment
         env.addFunc(name, tree.getSubTree(2));
         System.out.println( env + "\n" ); // for debugging purposes
      }
      else // not a function, so check the expression
      {
         // get the variable name
         String name = tree.getSubTree(1).getElement();

         // check the type of the expression
         String exprType = checkN( tree.getSubTree(2), env );

         if ( ! exprType.equals(declrType) )
         {
            System.out.println("Type error: " + AST2infix.ast2infix(tree) );
            typeChecked = false;
         }

         // check if this variable has already been defined
         if ( ! env.definedLocal(name) )
         {
            // add a "default" value to the environment
            env.add(name, new Value(declrType) );
         }
         else // this variable is already in the Environment object
         {
            // nothing needs to be done in this case
         }
         System.out.println( env + "\n" ); // for debugging purposes
      }
   }//checkDefine()


   /**
      Type check lambda expressions. Make sure the type of the body
      of the function agrees with the type of the return value found
      in the Signature.
   */
   private static boolean checkLambda(Tree tree, Environment env)
   {
      boolean result = true;
      // Retrive, from within the function's lambda expression,
      // the function's signature.
      Tree signature = tree.getSubTree(0);

      // Retrive, from within the function's signature,
      // the function's return type.
      String returnType = signature.getSubTree(0).getElement();

      // Retrive, from within the function's signature,
      // the formal parameter type.
      String formalType = signature.getSubTree(1).getElement();

      // Retrive, from within the function's lambda expression,
      // the formal parameter.
      String formalParam = tree.getSubTree(1).getElement();

      // Retrive, from within the function's lambda expression,
      // the body of the function.
      Tree expr = tree.getSubTree(2);

      // Create a "dummy" actual parameter of the correct type
      // with a default value.
      Value actualParam = new Value(formalType);

      // Create a local environment with the formal parameter bound
      // to the dummy actual parameter (that has the correct type).
      Environment localEnv = new Environment(env);  // NOTE: static scope
      localEnv.add(formalParam, actualParam );
      System.out.println( localEnv + "\n" ); // for debugging purposes

      // Check the type of the function's body (using the local
      // environment with the dummy binding).
      String exprType = checkE(expr, localEnv);

      // Check that the type of the function's body
      // agrees with the function's return type.
      if ( ! exprType.equals(returnType) )
      {
         result = false;
         typeChecked = false;
      }

      return result;
   }//checkLambda()


   private static String checkE(Tree tree, Environment env)
   {
      String result = null;

      String node = tree.getElement();

      if ( node.equals("begin") )
      {
         result = checkBegin( tree, env );
      }
      else if ( node.equals("var") )
      {
         result = checkVar( tree, env );
      }
      else
      {
         result = checkN( tree, env );
      }

      return result;
   }//checkE()


   private static String checkBegin(Tree tree, Environment env)
   {
      // create a new environment "nested" in the previous environment
      Environment localEnv = new Environment(env);

      String result = null;

      // check each expression (using the nested environment)
      for (int i = 0; i < tree.degree(); i++)
      {
         result = checkE( tree.getSubTree(i), localEnv );
      }

      return result;
   }//checkBegin()


   private static String checkVar(Tree tree, Environment env)
   {
      // get the type declaration
      String declrType = tree.getSubTree(0).getElement();

      // determine if the definition is a function
      if ( declrType.equals("func") )
      {
         // get the function name
         String name = tree.getSubTree(1).getElement();

         // check that the lambda expression is type correct
         if ( ! checkLambda( tree.getSubTree(2), env ) )
         {
            System.out.println("Type error: incorrect return type in " + name);
            typeChecked = false;
         }

         declrType = "lambda";

         // check if this variable has already been defined
         if ( ! env.definedLocalFunc(name) )
         {
            env.addFunc( name, tree.getSubTree(2) );
         }
         else // this variable is already in the local Environment object
         {
            System.out.println("Syntax error: local function already exisits: " + name);
            typeChecked = false;
         }
         System.out.println( env + "\n" ); // for debugging purposes
      }
      else // not a function
      {
         // get the variable name
         String name = tree.getSubTree(1).getElement();

         // check the type of the expression
         String exprType = checkN( tree.getSubTree(2), env );

         if ( ! exprType.equals(declrType) )
         {
            System.out.println("Type error: " + AST2infix.ast2infix(tree) );
            typeChecked = false;
         }

         // check if this variable has already been defined
         if ( ! env.definedLocal(name) )
         {
            // add a "default" value to the environment
            env.add(name, new Value(declrType) );
         }
         else // this variable is already in the local Environment object
         {
            System.out.println("Syntax error: local variable already exisits: " + name);
            typeChecked = false;
         }
         System.out.println( env + "\n" ); // for debugging purposes
      }

      return declrType;
   }//checkVar()


   private static String checkN(Tree tree, Environment env)
   {
      String result = null;

      String node = tree.getElement();

      if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = "bool";
         }
         else if ( node.matches("^[0-9][0-9]*") )
         {
            result = "int";
         }
         else if ( env.defined(node) )
         {
            // get this variable's type
            Value value = env.lookUp(node);
            result = value.tag;
         }
         else
         {
            System.out.println("Syntax error: undefined variable: " + node);
            result = "bool";  // a default type (this is not really a good idea?)
            typeChecked = false;
         }
      }
      else if ( node.equals("apply") )
      {
         result = checkApply( tree, env );
      }
      else if ( node.equals("set") )
      {
         result = checkSet( tree, env );
      }
      else if ( node.equals("print") )
      {
         result = checkPrint( tree, env );
      }
      else if ( node.equals("+") ||     // arithmetic operators
                node.equals("-") ||
                node.equals("*") ||
                node.equals("/") ||
                node.equals("%") ||
                node.equals("^") )
      {
         result = checkA( tree, env );
      }
      else if ( node.equals("&&") ||    // boolean operators
                node.equals("||") ||
                node.equals("!") )
      {
         result = checkB( tree, env );
      }
      else if ( node.equals("==") ||    // equality operators
                node.equals("!=") )
      {
         result = checkEq( tree, env );
      }
      else if ( node.equals("<") ||     // relational operators
                node.equals(">") ||
                node.equals("<=") ||
                node.equals(">=") )
      {
         result = checkR( tree, env );
      }
      else
      {
         System.out.println("Syntax error: invalid expression: " + AST2infix.ast2infix(tree));
         typeChecked = false;
      }

      return result;
   }//checkN()


   /**
      This method checks the application of a function to an actual parameter.
      This method checks that the declared type of the formal parameter matches
      the type the actual parameter in this application. The type returned by
      this method is the function's declared return type, found in the function's
      Signature.
   */
   private static String checkApply(Tree tree, Environment env)
   {
      // Check the type of the actual parmeter.
      String actualType = checkN( tree.getSubTree(1), env );

      String returnType;

      // Check if the function is defined
      String functionName = tree.getSubTree(0).getElement();
      if ( env.definedFunc( functionName ) )
      {
         // Retrive, from within the current environment,
         // a reference to the function's IPEP pair.
         IPEP ipepPair = env.lookUpFunc( functionName );

         // Retrive, from within the IPEP pair,
         // a reference to the function's definition
         // (i.e., the function's "lambda expression").
         Tree function = ipepPair.exp;

         // Retrive, from within the function's lambda expression,
         // the function's signature.
         Tree signature = function.getSubTree(0);

         // Retrive, from within the function's signature,
         // the function's return type.
         returnType = signature.getSubTree(0).getElement();

         // Retrive, from within the function's signature,
         // the formal parameter type.
         String formalType = signature.getSubTree(1).getElement();

         // Check that the types of the formal parameter
         // and the actual parameter agree.
         if ( ! formalType.equals(actualType) )
         {
            System.out.println("Type error: " + AST2infix.ast2infix(tree) );
            typeChecked = false;
         }
      }
      else // the function is not defined
      {
         System.out.println("Syntax error: undefined function: " + functionName);
         returnType = actualType;  // a reasonable guess at a return type
         typeChecked = false;
      }

      return returnType;
   }//checkApply()


   private static String checkSet(Tree tree, Environment env)
   {
      String varType = null;
      String exprType = null;

      // get the variable
      String variable = tree.getSubTree(0).getElement();

      // get, and then check, the expression
      Tree expr = tree.getSubTree(1);
      exprType = checkE( expr, env );

      // check if this variable has already been defined
      if ( ! env.defined(variable) )
      {
         System.out.println("Syntax error: undefined variable: " + variable);
         varType = exprType;  // just to keep things moving along
         typeChecked = false;
      }
      else // this variable is already in the environment
      {
         // get this variable's type
         Value value = env.lookUp(variable);
         varType = value.tag;
      }

      if ( ! varType.equals(exprType) )
      {
         System.out.println("Type error: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      return varType;
   }//checkSet()


   private static String checkPrint(Tree tree, Environment env)
   {
      return checkN( tree.getSubTree(0), env );
   }//checkPrint()


   // check an arithmetic expression
   private static String checkA(Tree tree, Environment env)
   {
      String typeL = checkN( tree.getSubTree(0), env );
      String typeR = checkN( tree.getSubTree(1), env );

      if ( ! typeL.equals("int") )
      {
         System.out.println("Type error: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }
      if ( ! typeR.equals("int") )
      {
         System.out.println("Type error: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      return "int";
   }//checkA()


   // check a boolean expression
   private static String checkB(Tree tree, Environment env)
   {
      String typeL = checkN( tree.getSubTree(0), env );
      if ( ! typeL.equals("bool") )
      {
         System.out.println("Type error: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      if ( tree.degree() == 2 )  // check for second operand
      {
         String typeR = checkN( tree.getSubTree(1), env );
         if ( ! typeR.equals("bool") )
         {
            System.out.println("Type error: " + AST2infix.ast2infix(tree) );
            typeChecked = false;
         }
      }

      return "bool";
   }//checkB()


   // check an equality expression (which is a kind of boolean expression)
   private static String checkEq(Tree tree, Environment env)
   {
      String typeL = checkE( tree.getSubTree(0), env );
      String typeR = checkE( tree.getSubTree(1), env );

      if ( ! typeL.equals(typeR) )
      {
         System.out.println("Type error: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      return "bool";
   }//checkEq()


   // check a relational expression (which is a kind of boolean expression)
   private static String checkR(Tree tree, Environment env)
   {
      String typeL = checkN( tree.getSubTree(0), env );
      String typeR = checkN( tree.getSubTree(1), env );

      if ( ! typeL.equals("int") )
      {
         System.out.println("Type error: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }
      if ( ! typeR.equals("int") )
      {
         System.out.println("Type error: " + AST2infix.ast2infix(tree) );
         typeChecked = false;
      }

      return "bool";
   }//checkR()

}//TypeChecker