/*
   The static method parse() is a recursive descent parser
   for the grammar in Language-2-infix.txt.

   This parser builds an abstract syntax tree. If an expression
   parses correctly, a reference to the AST is returned.
   If an expression does not parse, a ParseException gets thrown.

   Use the program TestParser.java to test this parser.
*/

public class Parser2AST
{
   public static Tree parse(String expression) throws ParseException
   {
      Tokenizer tokens = new Tokenizer(expression);  // Tokenizer is defined
                                                     // in Tokenizer.java
      Tree ast = parseExp(tokens);

      if ( tokens.hasToken() )  // there shouldn't be any more tokens
         throw new ParseException("unexpected input: "+tokens.peekToken()+"\n"+tokens);

      return ast;
   }//buildASTree()


   private static Tree parseExp(Tokenizer tokens) throws ParseException
   {
      Tree result = parseTerm(tokens);
      while ( tokens.hasToken()
            && (tokens.peekToken().equals("+")
             || tokens.peekToken().equals("-")) )
      {
         String operator = tokens.nextToken();
         Tree temp = new Tree( operator );
         temp.addSubTree( result );
         temp.addSubTree( parseTerm(tokens) );
         result = temp;
      }
      return result;
   }//parseExp()


   private static Tree parseTerm(Tokenizer tokens) throws ParseException
   {
      Tree result = parseSigned(tokens);
      while ( tokens.hasToken()
            && (tokens.peekToken().equals("*")
             || tokens.peekToken().equals("/")
             || tokens.peekToken().equals("%")) )
      {
         String operator = tokens.nextToken();
         Tree temp = new Tree( operator );
         temp.addSubTree( result );
         temp.addSubTree( parseSigned(tokens) );
         result = temp;
      }
      return result;
   }//parseTerm()


   private static Tree parseSigned(Tokenizer tokens) throws ParseException
   {
      if ( tokens.peekToken().equals("-") )
      {
         Tree temp = new Tree( tokens.nextToken() );
         temp.addSubTree( parseFactor(tokens) );
         return temp;
      }
      else
      {
         return parseFactor(tokens);
      }
   }//parseSigned()


   private static Tree parseFactor(Tokenizer tokens) throws ParseException
   {
      Tree result = parseBase(tokens);
      if ( tokens.hasToken() && tokens.peekToken().equals("^") )
      {
         Tree temp = new Tree( tokens.nextToken() );
         temp.addSubTree( result );
         temp.addSubTree( parseSigned(tokens) );
         result = temp;
      }
      return result;
   }//parseFactor()


   private static Tree parseBase(Tokenizer tokens) throws ParseException
   {
      Tree result = null;

      if ( tokens.peekToken().equals("(") )
      {
         tokens.match("(");
         result = parseExp(tokens);
         tokens.match(")");
      }
      else if ( tokens.peekToken().matches("^[0-9][0-9]*") )
      {
         result = new Tree( tokens.nextToken() );
      }
      else
      {
         throw new ParseException("incorrect base: " + tokens.nextToken());
      }
      return result;
   }//parseBase()

}//Parser2AST